import pathlib
from inspect import Parameter, Signature, signature
from io import IOBase
from typing import Optional, Dict, Callable, Union, IO, Any, Iterable, Tuple

import jinja2
import yaml

__all__ = ["cell_from_yaml_template"]


def split_default_settings_from_yaml(yaml_lines: Iterable[str]) -> Tuple[str, str]:
    """Separates out the 'default_settings' block from the rest of the file body.
    Note: 'default settings' MUST be at the TOP of the file.

    Args:
        yaml_lines: the lines of text in the yaml file

    Returns:
        a tuple of (main file contents), (setting block), both as multi-line strings
    """
    settings_lines = []
    other_lines = []
    # start reading all lines
    while yaml_lines:
        # pop lines until we find the default_settings block
        line = yaml_lines.pop(0)
        if line.startswith("default_settings"):
            settings_lines.append(line)
            # keep adding lines to settings until we find a new top-level block...
            # then we will add the rest of the lines to the main file block
            while yaml_lines:
                next_line = yaml_lines.pop(0)
                if next_line[0].isspace():
                    settings_lines.append(next_line)
                else:
                    other_lines.append(next_line)
                    break
        else:
            other_lines.append(line)
    settings_string = "\n".join(settings_lines)
    other_string = "\n".join(other_lines)
    return other_string, settings_string


def _split_yaml_definition(subpic_yaml):
    if isinstance(subpic_yaml, IOBase):
        f = subpic_yaml
        subpic_text = f.readlines()
    else:
        with open(subpic_yaml) as f:
            subpic_text = f.readlines()
    main_file, default_settings_string = split_default_settings_from_yaml(subpic_text)
    if default_settings_string:
        default_settings = yaml.safe_load(default_settings_string)["default_settings"]
    else:
        default_settings = {}
    return main_file, default_settings


def cell_from_yaml_template(
    filename: Union[str, IO[Any], pathlib.Path],
    name: str,
    routing_strategy: Optional[Dict[str, Callable]] = None,
) -> Callable:
    """Gets a PIC factory function from a yaml definition, which can optionally be a jinja template.

    Args:
        filename: the filepath of the pic yaml template
        name: the name of the component to create
        routing_strategy: a dictionary of routing functions
    Returns:
         a factory function for the component
    """
    from gdsfactory.pdk import get_routing_strategies

    if routing_strategy is None:
        routing_strategy = get_routing_strategies()
    return yaml_cell(
        yaml_definition=filename, name=name, routing_strategy=routing_strategy
    )


def get_default_settings_dict(default_settings):
    settings = {}
    for k, v in default_settings.items():
        try:
            settings[k] = v["value"]
        except TypeError as te:
            raise TypeError(
                f'Default setting "{k}" should be a dictionary with "value" defined.'
            ) from te
        except KeyError as ke:
            raise KeyError(
                f'Required key "value" not supplied for default setting "{k}"'
            ) from ke
    return settings


def yaml_cell(yaml_definition, name, routing_strategy):
    """The "cell" decorator equivalent for yaml files. Generates a proper cell function for yaml-defined circuits.

    Args:
        yaml_definition: the filename to the pic yaml definition
        name: the name of the pic to create
        routing_strategy: a dictionary of routing strategies to use for pic generation
    Returns:
        a dynamically-generated function for the yaml file
    """
    from gdsfactory.cell import cell_without_validator

    yaml_body, default_settings_def = _split_yaml_definition(yaml_definition)
    default_settings = get_default_settings_dict(default_settings_def)

    def _yaml_func(**kwargs):
        evaluated_text = _evaluate_yaml_template(yaml_body, default_settings, kwargs)
        return _pic_from_templated_yaml(evaluated_text, name, routing_strategy)

    sig = signature(_yaml_func)
    params = []
    docstring_lines = [
        f"{name}: a templated yaml cell. This cell accepts keyword arguments only",
        "",
    ]
    if default_settings:
        docstring_lines.append("Keyword Args:")

    for default_key, default_value in default_settings.items():
        p = Parameter(
            name=default_key, kind=Parameter.KEYWORD_ONLY, default=default_value
        )
        params.append(p)
        description = default_settings_def[default_key].get(
            "description", "No description given"
        )
        docstring_lines.append(f"    {default_key}: {description}")
    new_sig = Signature(parameters=params, return_annotation=sig.return_annotation)
    docstring = "\n".join(docstring_lines)

    _yaml_func.__name__ = name
    _yaml_func.__module__ = "yaml_jinja"
    _yaml_func.__signature__ = new_sig
    _yaml_func.__doc__ = docstring
    return cell_without_validator(_yaml_func)


def _evaluate_yaml_template(main_file, default_settings, settings):
    template = jinja2.Template(main_file)
    complete_settings = dict(default_settings)
    complete_settings.update(settings)
    return template.render(**complete_settings)


def _pic_from_templated_yaml(evaluated_text, name, routing_strategy):
    """Creates a component from a  *.pic.yml file. This is a lower-level function. See from_yaml_template() for more common usage.

    Args:
        name: the pic name
        routing_strategy: a dictionary of route factories
    Returns: the component
    """
    from gdsfactory.read.from_yaml import from_yaml

    c = from_yaml(
        evaluated_text,
        routing_strategy=routing_strategy,
    ).copy()
    c.name = name
    return c



import gdsfactory as gf
from gdsfactory.simulation.gmeep.write_sparameters_meep import write_sparameters_meep


def test_materials_override() -> None:
    """Checks that materials are properly overridden if index is provided."""

    c = gf.components.straight(length=2)

    # Default (materials strings)
    sp1 = write_sparameters_meep(
        c,
        run=False,
        animate=False,
        is_3d=False,
        overwrite=True,
    )

    # Override
    sp2 = write_sparameters_meep(
        c,
        run=False,
        animate=False,
        material_name_to_meep=dict(si=2.0),
        is_3d=False,
        overwrite=True,
    )

    assert sp1.geometry[0].material.epsilon(freq=1)[0][0] != 4.0
    assert sp2.geometry[0].material.epsilon(freq=1)[0][0] == 4.0


if __name__ == "__main__":
    test_materials_override()



from __future__ import annotations

from types import LambdaType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import nlopt
import numpy as np
from meep import Block, EigenModeSource, MaterialGrid, Simulation, Vector3, Volume
from meep.adjoint import DesignRegion, EigenmodeCoefficient, OptimizationProblem
from meep.visualization import get_2D_dimensions
from numpy import ndarray

import gdsfactory as gf
from gdsfactory import Component
from gdsfactory.simulation.gmeep import get_simulation
from gdsfactory.technology import LayerStack
from gdsfactory.typings import Layer


def get_meep_adjoint_optimizer(
    component: Component,
    objective_function: Callable,
    design_regions: List[DesignRegion],
    design_variables: List[MaterialGrid],
    design_update: np.ndarray,
    TE_mode_number: int = 1,
    resolution: int = 30,
    cell_size: Optional[Tuple] = None,
    extend_ports_length: Optional[float] = 10.0,
    layer_stack: Optional[LayerStack] = None,
    zmargin_top: float = 3.0,
    zmargin_bot: float = 3.0,
    tpml: float = 1.5,
    clad_material: str = "SiO2",
    is_3d: bool = False,
    wavelength_start: float = 1.5,
    wavelength_stop: float = 1.6,
    wavelength_points: int = 50,
    dfcen: float = 0.2,
    port_source_name: str = "o1",
    port_margin: float = 3,
    distance_source_to_monitors: float = 0.2,
    port_source_offset: float = 0,
    port_monitor_offset: float = 0,
    dispersive: bool = False,
    material_name_to_meep: Optional[Dict[str, Union[str, float]]] = None,
    **settings,
):
    """Return a Meep `OptimizationProblem` object.

    Args:
        component: gdsfactory component.
        objective_function: functions must be composed of "field functions" that transform the recorded fields.
        design_regions: list of DesignRegion objects.
        design_variables: list of MaterialGrid objects.
        design_update: ndarray to initialize the optimization.
        TE_mode_number: TE mode number.
        resolution: in pixels/um (20: for coarse, 120: for fine).
        cell_size: tuple of Simulation object dimensions in um.
        extend_ports_length: to extend ports beyond the PML.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        zmargin_top: thickness for cladding above core.
        zmargin_bot: thickness for cladding below core.
        tpml: PML thickness (um).
        clad_material: material for cladding.
        is_3d: if True runs in 3D.
        wavelength_start: wavelength min (um).
        wavelength_stop: wavelength max (um).
        wavelength_points: wavelength steps.
        dfcen: delta frequency.
        port_source_name: input port name.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before.
        port_source_offset: offset between source GDS port and source MEEP port.
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port.
        dispersive: use dispersive material models (requires higher resolution).
        material_name_to_meep: map layer_stack names with meep material database name
            or refractive index. dispersive materials have a wavelength dependent index.

    Keyword Args:
        settings: extra simulation settings (resolution, symmetries, etc.)

    Returns:
        opt: OptimizationProblem object
    """
    sim_dict = get_simulation(
        component,
        resolution=resolution,
        extend_ports_length=extend_ports_length,
        layer_stack=layer_stack,
        zmargin_top=zmargin_top,
        zmargin_bot=zmargin_bot,
        tpml=tpml,
        clad_material=clad_material,
        is_3d=is_3d,
        wavelength_start=wavelength_start,
        wavelength_stop=wavelength_stop,
        wavelength_points=wavelength_points,
        dfcen=dfcen,
        port_source_name=port_source_name,
        port_margin=port_margin,
        distance_source_to_monitors=distance_source_to_monitors,
        port_source_offset=port_source_offset,
        port_monitor_offset=port_monitor_offset,
        dispersive=dispersive,
        material_name_to_meep=material_name_to_meep,
        **settings,
    )
    sim = sim_dict["sim"]

    design_regions_geoms = [
        Block(
            center=design_region.center,
            size=design_region.size,
            material=design_variable,
        )
        for design_region, design_variable in zip(design_regions, design_variables)
    ]

    for design_region_geom in design_regions_geoms:
        sim.geometry.append(design_region_geom)

    cell_thickness = sim.cell_size[2]

    monitors = sim_dict["monitors"]
    ob_list = [
        EigenmodeCoefficient(
            sim,
            Volume(
                center=monitor.regions[0].center,
                size=monitor.regions[0].size,
            ),
            TE_mode_number,
        )
        for monitor in monitors.values()
    ]

    c = component.copy()
    for design_region, design_variable in zip(design_regions, design_variables):
        sim.geometry.append(
            Block(design_region.size, design_region.center, material=design_variable)
        )
        block = c << gf.components.rectangle(
            (design_region.size[0], design_region.size[1])
        )
        block.center = (design_region.center[0], design_region.center[1])

    sim.cell_size = (
        Vector3(*cell_size)
        if cell_size
        else Vector3(
            c.xsize + 2 * sim.boundary_layers[0].thickness,
            c.ysize + 2 * sim.boundary_layers[0].thickness,
            cell_thickness,
        )
    )

    source = [
        EigenModeSource(
            sim.sources[0].src,
            eig_band=1,
            direction=sim.sources[0].direction,
            eig_kpoint=Vector3(1, 0, 0),
            size=sim.sources[0].size,
            center=sim.sources[0].center,
        )
    ]

    sim.sources = source

    opt = OptimizationProblem(
        simulation=sim,
        objective_functions=[objective_function],
        objective_arguments=ob_list,
        design_regions=design_regions,
        frequencies=sim_dict["freqs"],
        decay_by=settings.get("decay_by", 1e-5),
    )

    opt.update_design([design_update])
    opt.plot2D(True)

    return opt


def run_meep_adjoint_optimizer(
    number_of_params: int,
    cost_function: LambdaType,
    update_variable: np.ndarray,
    maximize_cost_function: bool = True,
    algorithm: int = nlopt.LD_MMA,
    lower_bound: Any = 0,
    upper_bound: Any = 1,
    maxeval: int = 10,
    get_optimized_component: bool = False,
    opt: OptimizationProblem = None,
    **kwargs,
) -> Union[ndarray, Component]:
    """Run adjoint optimization using Meep.

    Args:
        number_of_params: number of parameters to optimize (usually resolution_in_x * resolution_in_y).
        cost_function: cost function to optimize.
        update_variable: variable to update the optimization with.
        maximize_cost_function: if True, maximize the cost function, else minimize it.
        algorithm: nlopt algorithm to use (default: nlopt.LD_MMA).
        lower_bound: lower bound for the optimization.
        upper_bound: upper bound for the optimization.
        maxeval: maximum number of evaluations.
        get_optimized_component: if True, returns the optimized gdsfactory Component.
            If this is True, the Optimization object used for the optimization must be passed as an argument.
        opt: OptimizationProblem object used for the optimization. Used only if get_optimized_component is True.

    Keyword Args:
        fcen: center frequency of the source.
        upscale_factor: upscale factor for the optimization's grid.
        threshold_offset_from_max: threshold offset from max eps value.
        layer: layer to apply to the optimized component.
    """
    solver = nlopt.opt(algorithm, number_of_params)
    solver.set_lower_bounds(lower_bound)
    solver.set_upper_bounds(upper_bound)
    if maximize_cost_function:
        solver.set_max_objective(cost_function)
    else:
        solver.set_min_objective(cost_function)
    solver.set_maxeval(maxeval)
    update_variable[:] = solver.optimize(update_variable)

    if get_optimized_component:
        fcen = kwargs.get("fcen", 1 / 1.55)
        upscale_factor = kwargs.get("upscale_factor", 2)
        threshold_offset_from_max = kwargs.get("threshold_offset_from_max", 0.01)
        layer = kwargs.get("layer", (1, 0))

        return get_component_from_sim(
            opt.sim, fcen, upscale_factor, threshold_offset_from_max, layer
        )
    return update_variable


def get_component_from_sim(
    sim: Simulation,
    fcen: float = 1 / 1.55,
    upscale_factor: int = 2,
    threshold_offset_from_max: float = 2.0,
    layer: Layer = (1, 0),
) -> Component:
    """Get gdsfactory Component from Meep Simulation object.

    Args:
        sim: Meep Simulation object.
        fcen: center frequency of the source.
        upscale_factor: upscale factor for the optimization's grid.
        threshold_offset_from_max: threshold offset from max eps value.
        layer: layer to apply to the optimized component.

    Returns:
        gdsfactory Component.
    """
    grid_resolution = upscale_factor * sim.resolution
    sim_center, sim_size = get_2D_dimensions(sim, output_plane=None)
    xmin = sim_center.x - sim_size.x / 2
    xmax = sim_center.x + sim_size.x / 2
    ymin = sim_center.y - sim_size.y / 2
    ymax = sim_center.y + sim_size.y / 2
    Nx = int((xmax - xmin) * grid_resolution + 1)
    Ny = int((ymax - ymin) * grid_resolution + 1)
    xtics = np.linspace(xmin, xmax, Nx)
    ytics = np.linspace(ymin, ymax, Ny)
    ztics = np.array([sim_center.z])
    eps_data = np.real(sim.get_epsilon_grid(xtics, ytics, ztics, frequency=fcen))
    return gf.read.from_np(
        eps_data,
        nm_per_pixel=1e3 / grid_resolution,
        layer=layer,
        threshold=np.max(eps_data) - threshold_offset_from_max,
    )


def _example_optim_geometry() -> Component:
    """Dummy example of a component to optimize."""
    from meep import Medium

    design_region_width = 5
    design_region_height = 4

    resolution = 20
    design_region_resolution = int(5 * resolution)

    Nx = int(design_region_resolution * design_region_width)
    Ny = int(design_region_resolution * design_region_height)

    pml_size = 1.0
    waveguide_length = 0.5
    Sx = 2 * pml_size + 2 * waveguide_length + design_region_width

    SiO2 = Medium(index=1.44)
    Si = Medium(index=3.4)

    design_variables = MaterialGrid(Vector3(Nx, Ny), SiO2, Si, grid_type="U_MEAN")
    design_region = DesignRegion(
        design_variables,
        volume=Volume(
            center=Vector3(),
            size=Vector3(design_region_width, design_region_height, 0),
        ),
    )

    c = Component("mmi1x2")

    arm_separation = 1.0
    straight1 = c << gf.components.straight(Sx / 2 + 1)
    straight1.move(straight1.ports["o2"], (-design_region_width / 2.0, 0))
    straight2 = c << gf.components.straight(Sx / 2 + 1)
    straight2.move(
        straight2.ports["o1"], (design_region_width / 2.0, (arm_separation + 1.0) / 2.0)
    )
    straight3 = c << gf.components.straight(Sx / 2 + 1)
    straight3.move(
        straight3.ports["o1"],
        (design_region_width / 2.0, (-arm_separation - 1.0) / 2.0),
    )

    c.add_port("o1", port=straight1.ports["o1"])
    c.add_port("o2", port=straight2.ports["o2"])
    c.add_port("o3", port=straight3.ports["o2"])

    return design_region, design_variables, c, Nx, Ny


if __name__ == "__main__":
    import autograd.numpy as npa

    eta_i = 0.5

    design_region, design_variables, c, Nx, Ny = _example_optim_geometry()

    seed = 240
    np.random.seed(seed)
    x0 = np.random.rand(
        Nx * Ny,
    )

    def J(source, top, bottom):
        power = npa.abs(top / source) ** 2 + npa.abs(bottom / source) ** 2
        return npa.mean(power)

    opt = get_meep_adjoint_optimizer(
        c,
        J,
        [design_region],
        [design_variables],
        x0,
        cell_size=(15, 8),
        extend_ports_length=0,
        port_margin=0.75,
        port_source_offset=-3.5,
        port_monitor_offset=-3.5,
    )

    opt.plot2D(True)



from gdsfactory.simulation.sax.parameter import LithoParameter
import gdsfactory as gf


def test_litho_parameters():
    c = gf.Component("myComponent")
    c.add_polygon(
        [[2.8, 3], [5, 3], [5, 0.8]],
        layer=1,
    )
    c.add_polygon(
        [
            [2, 0],
            [2, 2],
            [4, 2],
            [4, 0],
        ],
        layer=1,
    )
    c.add_polygon(
        [
            [0, 0.5],
            [0, 1.5],
            [3, 1.5],
            [3, 0.5],
        ],
        layer=1,
    )
    c.add_polygon(
        [
            [0, 0],
            [5, 0],
            [5, 3],
            [0, 3],
        ],
        layer=2,
    )
    c.add_polygon(
        [
            [2.5, -2],
            [3.5, -2],
            [3.5, -0.1],
            [2.5, -0.1],
        ],
        layer=1,
    )
    c.add_port(name="o1", center=(0, 1), width=1, orientation=0, layer=1)
    c.add_port(name="o2", center=(3, -2), width=1, orientation=90, layer=1)

    param = LithoParameter(layername="core")
    param.layer_dilation_erosion(c, 0.2)

    param = LithoParameter(layername="core")
    param.layer_dilation_erosion(c, -0.2)

    param = LithoParameter(layername="core")
    param.layer_x_offset(c, 0.5)

    param = LithoParameter(layername="core")
    param.layer_y_offset(c, 0.5)

    param = LithoParameter(layername="core")
    param.layer_round_corners(c, 0.2)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.typings import Component, ComponentSpec, Optional, Tuple

_cols_200mm_wafer = (2, 6, 6, 8, 8, 6, 6, 2)


@gf.cell
def wafer(
    reticle: ComponentSpec = "die",
    cols: Tuple[int, ...] = _cols_200mm_wafer,
    xspacing: Optional[float] = None,
    yspacing: Optional[float] = None,
) -> Component:
    """Returns complete wafer. Useful for mask aligner steps.

    Args:
        reticle: spec for each wafer reticle.
        cols: how many columns per row.
        xspacing: optional spacing, defaults to reticle.xsize.
        yspacing: optional spacing, defaults to reticle.ysize.
    """
    c = gf.Component()
    reticle = gf.get_component(reticle)
    xspacing = xspacing or reticle.xsize
    yspacing = yspacing or reticle.ysize

    for i, columns in enumerate(cols):
        ref = c.add_array(
            reticle, rows=1, columns=columns, spacing=(xspacing, yspacing)
        )
        ref.x = 0
        ref.movey(i * yspacing)

    return c


if __name__ == "__main__":
    c = wafer()
    c.show()



"""PDK stores layers, cross_sections, cell functions ..."""

from __future__ import annotations
import pathlib
import warnings
from functools import partial
from typing import Any, Callable, Optional, Union, Tuple
from typing_extensions import Literal

import numpy as np
from omegaconf import DictConfig
from pydantic import BaseModel, Field, validator

from gdsfactory.config import PATH, logger
from gdsfactory.containers import containers as containers_default
from gdsfactory.events import Event
from gdsfactory.materials import MaterialSpec
from gdsfactory.materials import materials_index as materials_index_default
from gdsfactory.read import cell_from_yaml
from gdsfactory.show import show
from gdsfactory.symbols import floorplan_with_block_letters
from gdsfactory.technology import LayerStack, LayerViews
from gdsfactory.typings import (
    CellSpec,
    Component,
    ComponentFactory,
    ComponentSpec,
    CrossSection,
    Transition,
    CrossSectionFactory,
    CrossSectionSpec,
    Dict,
    Layer,
    LayerSpec,
    PathType,
)

component_settings = ["function", "component", "settings"]
cross_section_settings = ["function", "cross_section", "settings"]
layers_required = []

constants = {
    "fiber_array_spacing": 127.0,
    "fiber_spacing": 50.0,
    "fiber_input_to_output_spacing": 200.0,
    "metal_spacing": 10.0,
    "pad_spacing": 100.0,
    "pad_size": (80, 80),
}

nm = 1e-3


class GdsWriteSettings(BaseModel):
    """Settings to use when writing to GDS."""

    on_uncached_component: Literal["warn", "error", "ignore"] = "ignore"
    unit: float = Field(
        default=1e-6,
        description="The units of coordinates in the database. The default is 1e-6 (1 micron).",
    )
    precision: float = Field(
        default=1e-9,
        description="The maximum precision of points in the database. For example, a value of 1e-9 would mean that you have a 1nm grid",
    )
    on_duplicate_cell: str = Field(
        default="warn",
        description="""Action to take when a duplicate cell is encountered on gds write (usually problematic). Options are
                        "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                        "error": throw a ValueError when attempting to write a gds with duplicate cells.
                        "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.""",
    )
    flatten_invalid_refs: bool = Field(
        default=False,
        description="If true, will auto-correct (and flatten) cell references which are off-grid or rotated by non-manhattan angles.",
    )
    max_points: int = Field(
        default=4000,
        description="Maximum number of points to allow in a polygon before fracturing.",
    )


class OasisWriteSettings(BaseModel):
    compression_level: int = Field(
        default=6,
        description="Level of compression for cells (between 0 and 9). Setting to 0 will disable cell compression, 1 gives the best speed, and 9 gives the best compression.",
    )
    detect_rectangles: bool = Field(
        default=True, description="If true, stores rectangles in a compressed format."
    )
    detect_trapezoids: bool = Field(
        default=True, description="If true, stores trapezoids in a compressed format."
    )
    circle_tolerance: bool = Field(
        default=0,
        description="Tolerance for detecting circles. If less or equal to 0, no detection is performed. Circles are stored in compressed format.",
    )
    validation: Optional[str] = Field(
        default=None,
        description="Type of validation to include in the saved file ('crc32', 'checksum32', or None).",
    )
    standard_properties: bool = Field(
        default=False,
        description="If true, stores standard OASIS properties in the file.",
    )


class Pdk(BaseModel):
    """Store layers, cross_sections, cell functions, simulation_settings ...

    only one Pdk can be active at a given time.

    Parameters:
        name: PDK name.
        cross_sections: dict of cross_sections factories.
        cells: dict of parametric cells that return Components.
        symbols: dict of symbols names to functions.
        default_symbol_factory:
        containers: dict of pcells that contain other cells.
        base_pdk: a pdk to copy from and extend.
        default_decorator: decorate all cells, if not otherwise defined on the cell.
        layers: maps name to gdslayer/datatype.
            For example dict(si=(1, 0), sin=(34, 0)).
        layer_stack: maps name to layer numbers, thickness, zmin, sidewall_angle.
            if can also contain material properties
            (refractive index, nonlinear coefficient, sheet resistance ...).
        layer_views: includes layer name to color, opacity and pattern.
        layer_transitions: transitions between different cross_sections.
        sparameters_path: to store Sparameters simulations.
        modes_path: to store Sparameters simulations.
        interconnect_cml_path: path to interconnect CML (optional).
        warn_off_grid_ports: raises warning when extruding paths with offgrid ports.
        constants: dict of constants for the PDK.
        materials_index: material spec names to material spec, which can be:
            string: material name.
            float: refractive index.
            float, float: refractive index real and imaginary part.
            function: function of wavelength.
        routing_strategies: functions enabled to route.
        circuit_yaml_parser: can parse different YAML formats.
        gds_write_settings: to write GDSII files.
        oasis_settings: to write OASIS files.
        bend_points_distance: default points distance for bends in um.

    """

    name: str
    cross_sections: Dict[str, CrossSectionFactory] = Field(default_factory=dict)
    cells: Dict[str, ComponentFactory] = Field(default_factory=dict)
    symbols: Dict[str, ComponentFactory] = Field(default_factory=dict)
    default_symbol_factory: Callable = floorplan_with_block_letters
    containers: Dict[str, ComponentFactory] = containers_default
    base_pdk: Optional[Pdk] = None
    default_decorator: Optional[Callable[[Component], None]] = None
    layers: Dict[str, Layer] = Field(default_factory=dict)
    layer_stack: Optional[LayerStack] = None
    layer_views: Optional[LayerViews] = None
    layer_transitions: Dict[Union[Layer, Tuple[Layer, Layer]], ComponentSpec] = Field(
        default_factory=dict
    )
    sparameters_path: Optional[PathType] = None
    modes_path: Optional[PathType] = PATH.modes
    interconnect_cml_path: Optional[PathType] = None
    warn_off_grid_ports: bool = False
    constants: Dict[str, Any] = constants
    materials_index: Dict[str, MaterialSpec] = materials_index_default
    routing_strategies: Optional[Dict[str, Callable]] = None
    circuit_yaml_parser: Callable = cell_from_yaml
    gds_write_settings: GdsWriteSettings = GdsWriteSettings()
    oasis_settings: OasisWriteSettings = OasisWriteSettings()
    bend_points_distance: float = 20 * nm

    @property
    def grid_size(self):
        """The minimum unit resolvable on the layout grid, relative to the unit."""
        return self.gds_write_settings.precision / self.gds_write_settings.unit

    @grid_size.setter
    def grid_size(self, value):
        self.gds_write_settings.precision = value * self.gds_write_settings.unit

    class Config:
        """Configuration."""

        extra = "forbid"
        fields = {
            "cross_sections": {"exclude": True},
            "cells": {"exclude": True},
            "containers": {"exclude": True},
            "default_symbol_factory": {"exclude": True},
            "default_decorator": {"exclude": True},
            "materials_index": {"exclude": True},
            "circuit_yaml_parser": {"exclude": True},
        }

    @validator("sparameters_path")
    def is_pathlib_path(cls, path):
        return pathlib.Path(path)

    def validate_layers(self):
        for layer in layers_required:
            if layer not in self.layers:
                raise ValueError(
                    f"{layer!r} not in Pdk.layers {list(self.layers.keys())}"
                )

    def activate(self) -> None:
        """Set current pdk to as the active pdk."""
        from gdsfactory.cell import clear_cache

        logger.info(f"{self.name!r} PDK is now active")

        clear_cache()

        if self.base_pdk:
            cross_sections = self.base_pdk.cross_sections
            cross_sections.update(self.cross_sections)
            self.cross_sections = cross_sections

            cells = self.base_pdk.cells
            cells.update(self.cells)
            self.cells.update(cells)

            containers = self.base_pdk.containers
            containers.update(self.containers)
            self.containers.update(containers)

            layers = self.base_pdk.layers
            layers.update(self.layers)
            self.layers.update(layers)

            if not self.default_decorator:
                self.default_decorator = self.base_pdk.default_decorator
        self.validate_layers()
        _set_active_pdk(self)

    def register_cells(self, **kwargs) -> None:
        """Register cell factories."""
        for name, cell in kwargs.items():
            if not callable(cell):
                raise ValueError(
                    f"{cell} is not callable, make sure you register "
                    "cells functions that return a Component"
                )
            if name in self.cells:
                warnings.warn(f"Overwriting cell {name!r}")

            self.cells[name] = cell
            on_cell_registered.fire(name=name, cell=cell, pdk=self)

    def register_containers(self, **kwargs) -> None:
        """Register container factories."""
        for name, cell in kwargs.items():
            if not callable(cell):
                raise ValueError(
                    f"{cell} is not callable, make sure you register "
                    "cells functions that return a Component"
                )
            if name in self.containers:
                warnings.warn(f"Overwriting container {name!r}")

            self.containers[name] = cell
            on_container_registered.fire(name=name, cell=cell, pdk=self)

    def register_cross_sections(self, **kwargs) -> None:
        """Register cross_sections factories."""
        for name, cross_section in kwargs.items():
            if not callable(cross_section):
                raise ValueError(
                    f"{cross_section} is not callable, make sure you register "
                    "cross_section functions that return a CrossSection"
                )
            if name in self.cross_sections:
                warnings.warn(f"Overwriting cross_section {name!r}")
            self.cross_sections[name] = cross_section
            on_cross_section_registered.fire(
                name=name, cross_section=cross_section, pdk=self
            )

    def register_cells_yaml(
        self,
        dirpath: Optional[PathType] = None,
        update: bool = False,
        **kwargs,
    ) -> None:
        """Load *.pic.yml YAML files and register them as cells.

        Args:
            dirpath: directory to recursive search for YAML cells.
            update: does not raise ValueError if cell already registered.

        Keyword Args:
            cell_name: cell function. To update cells dict.

        """

        message = "Updated" if update else "Registered"

        if dirpath:
            dirpath = pathlib.Path(dirpath)

            if not dirpath.is_dir():
                raise ValueError(f"{dirpath!r} needs to be a directory.")

            for filepath in dirpath.glob("*/**/*.pic.yml"):
                name = filepath.stem.split(".")[0]
                if not update and name in self.cells:
                    raise ValueError(
                        f"ERROR: Cell name {name!r} from {filepath} already registered."
                    )
                parser = self.circuit_yaml_parser
                self.cells[name] = parser(filepath, name=name)
                on_yaml_cell_registered.fire(name=name, cell=self.cells[name], pdk=self)
                logger.info(f"{message} cell {name!r}")

        for k, v in kwargs.items():
            if not update and k in self.cells:
                raise ValueError(f"ERROR: Cell name {k!r} already registered.")
            self.cells[k] = v
            logger.info(f"{message} cell {k!r}")

    def remove_cell(self, name: str):
        """Removes cell from a PDK."""
        if name not in self.cells:
            raise ValueError(f"{name!r} not in {list(self.cells.keys())}")
        self.cells.pop(name)
        logger.info(f"Removed cell {name!r}")

    def get_cell(self, cell: CellSpec, **kwargs) -> ComponentFactory:
        """Returns ComponentFactory from a cell spec."""
        cells_and_containers = set(self.cells.keys()).union(set(self.containers.keys()))

        if callable(cell):
            return cell
        elif isinstance(cell, str):
            if cell not in cells_and_containers:
                cells = list(self.cells.keys())
                containers = "\n".join(list(self.containers.keys()))
                raise ValueError(
                    f"{cell!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = self.cells[cell] if cell in self.cells else self.containers[cell]
            return cell
        elif isinstance(cell, (dict, DictConfig)):
            for key in cell.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(cell.get("settings", {}))
            settings.update(**kwargs)

            cell_name = cell.get("function")
            if not isinstance(cell_name, str) or cell_name not in cells_and_containers:
                cells = list(self.cells.keys())
                containers = list(self.containers.keys())
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = (
                self.cells[cell_name]
                if cell_name in self.cells
                else self.containers[cell_name]
            )
            return partial(cell, **settings)
        else:
            raise ValueError(
                "get_cell expects a CellSpec (ComponentFactory, string or dict),"
                f"got {type(cell)}"
            )

    def get_component(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns component from a component spec."""
        return self._get_component(
            component=component, cells=self.cells, containers=self.containers, **kwargs
        )

    def get_symbol(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns a component's symbol from a component spec."""
        # this is a pretty rough first implementation
        try:
            self._get_component(
                component=component, cells=self.symbols, containers={}, **kwargs
            )
        except ValueError:
            component = self.get_component(component, **kwargs)
            return self.default_symbol_factory(component)

    def _get_component(
        self,
        component: ComponentSpec,
        cells: Dict[str, Callable],
        containers: Dict[str, Callable],
        **kwargs,
    ) -> Component:
        """Returns component from a component spec."""
        cells_and_containers = set(cells.keys()).union(set(containers.keys()))

        if isinstance(component, Component):
            if kwargs:
                raise ValueError(f"Cannot apply kwargs {kwargs} to {component.name!r}")
            return component
        elif callable(component):
            return component(**kwargs)
        elif isinstance(component, str):
            if component not in cells_and_containers:
                cells = list(cells.keys())
                containers = list(containers.keys())
                raise ValueError(
                    f"{component!r} not in PDK {self.name!r} cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = cells[component] if component in cells else containers[component]
            return cell(**kwargs)
        elif isinstance(component, (dict, DictConfig)):
            for key in component.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(component.get("settings", {}))
            settings.update(**kwargs)

            cell_name = component.get("component", None)
            cell_name = cell_name or component.get("function")
            if not isinstance(cell_name, str) or cell_name not in cells_and_containers:
                cells = list(cells.keys())
                containers = list(containers.keys())
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = cells[cell_name] if cell_name in cells else containers[cell_name]
            component = cell(**settings)
            return component
        else:
            raise ValueError(
                "get_component expects a ComponentSpec (Component, ComponentFactory, "
                f"string or dict), got {type(component)}"
            )

    def get_cross_section(
        self, cross_section: CrossSectionSpec, **kwargs
    ) -> Union[CrossSection, Transition]:
        """Returns cross_section from a cross_section spec."""
        if isinstance(cross_section, CrossSection):
            return cross_section.copy(**kwargs)
        elif isinstance(cross_section, Transition):
            return cross_section.copy(**kwargs)
        elif callable(cross_section):
            return cross_section(**kwargs)
        elif isinstance(cross_section, str):
            if cross_section not in self.cross_sections:
                cross_sections = list(self.cross_sections.keys())
                raise ValueError(f"{cross_section!r} not in {cross_sections}")
            cross_section_factory = self.cross_sections[cross_section]
            return cross_section_factory(**kwargs)
        elif isinstance(cross_section, (dict, DictConfig)):
            for key in cross_section.keys():
                if key not in cross_section_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {cross_section_settings}"
                    )
            cross_section_factory_name = cross_section.get("cross_section", None)
            cross_section_factory_name = (
                cross_section_factory_name or cross_section.get("function")
            )
            if (
                not isinstance(cross_section_factory_name, str)
                or cross_section_factory_name not in self.cross_sections
            ):
                cross_sections = list(self.cross_sections.keys())
                raise ValueError(
                    f"{cross_section_factory_name!r} not in {cross_sections}"
                )
            cross_section_factory = self.cross_sections[cross_section_factory_name]
            settings = dict(cross_section.get("settings", {}))
            settings.update(**kwargs)

            return cross_section_factory(**settings)
        else:
            raise ValueError(
                "get_cross_section expects a CrossSectionSpec (CrossSection, "
                f"CrossSectionFactory, Transition, string or dict), got {type(cross_section)}"
            )

    def get_layer(self, layer: LayerSpec) -> Layer:
        """Returns layer from a layer spec."""
        if isinstance(layer, (tuple, list)):
            if len(layer) != 2:
                raise ValueError(f"{layer!r} needs two integer numbers.")
            return layer
        elif isinstance(layer, int):
            return (layer, 0)
        elif isinstance(layer, str):
            if layer not in self.layers:
                raise ValueError(f"{layer!r} not in {self.layers.keys()}")
            return self.layers[layer]
        elif layer is np.nan:
            return np.nan
        elif layer is None:
            return
        else:
            raise ValueError(
                f"{layer!r} needs to be a LayerSpec (string, int or Layer)"
            )

    def get_layer_views(self) -> LayerViews:
        if self.layer_views is None:
            raise ValueError(f"layer_views for Pdk {self.name!r} is None")
        return self.layer_views

    def get_layer_stack(self) -> LayerStack:
        if self.layer_stack is None:
            raise ValueError(f"layer_stack for Pdk {self.name!r} is None")
        return self.layer_stack

    def get_constant(self, key: str) -> Any:
        if not isinstance(key, str):
            return key
        if key not in self.constants:
            constants = list(self.constants.keys())
            raise ValueError(f"{key!r} not in {constants}")
        return self.constants[key]

    def get_material_index(self, key: str, *args, **kwargs) -> float:
        if key not in self.materials_index:
            material_names = list(self.materials_index.keys())
            raise ValueError(f"{key!r} not in {material_names}")
        material = self.materials_index[key]
        return material(*args, **kwargs) if callable(material) else material

    # _on_cell_registered = Event()
    # _on_container_registered: Event = Event()
    # _on_yaml_cell_registered: Event = Event()
    # _on_cross_section_registered: Event = Event()
    #
    # @property
    # def on_cell_registered(self) -> Event:
    #     return self._on_cell_registered
    #
    # @property
    # def on_container_registered(self) -> Event:
    #     return self._on_container_registered
    #
    # @property
    # def on_yaml_cell_registered(self) -> Event:
    #     return self._on_yaml_cell_registered
    #
    # @property
    # def on_cross_section_registered(self) -> Event:
    #     return self._on_cross_section_registered


_ACTIVE_PDK = None


def get_active_pdk() -> Pdk:
    global _ACTIVE_PDK
    if _ACTIVE_PDK is None:
        logger.warning("No active PDK. Activating generic PDK.\n")
        import gdsfactory as gf

        PDK = gf.get_generic_pdk()
        PDK.activate()
        _ACTIVE_PDK = PDK
    return _ACTIVE_PDK


def get_material_index(material: MaterialSpec, *args, **kwargs) -> Component:
    return get_active_pdk().get_material_index(material, *args, **kwargs)


def get_component(component: ComponentSpec, **kwargs) -> Component:
    return get_active_pdk().get_component(component, **kwargs)


def get_cell(cell: CellSpec, **kwargs) -> ComponentFactory:
    return get_active_pdk().get_cell(cell, **kwargs)


def get_cross_section(cross_section: CrossSectionSpec, **kwargs) -> CrossSection:
    return get_active_pdk().get_cross_section(cross_section, **kwargs)


def get_layer(layer: LayerSpec) -> Layer:
    return get_active_pdk().get_layer(layer)


def get_layer_views() -> LayerViews:
    return get_active_pdk().get_layer_views()


def get_layer_stack() -> LayerStack:
    return get_active_pdk().get_layer_stack()


def get_grid_size() -> float:
    return get_active_pdk().grid_size


def get_constant(constant_name: Any) -> Any:
    """If constant_name is a string returns a the value from the dict."""
    return (
        get_active_pdk().get_constant(constant_name)
        if isinstance(constant_name, str)
        else constant_name
    )


def get_sparameters_path() -> pathlib.Path:
    PDK = get_active_pdk()
    if PDK.sparameters_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no sparameters_path")
    return PDK.sparameters_path


def get_modes_path() -> Optional[pathlib.Path]:
    PDK = get_active_pdk()
    return PDK.modes_path


def get_interconnect_cml_path() -> pathlib.Path:
    PDK = get_active_pdk()
    if PDK.interconnect_cml_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no interconnect_cml_path")
    return PDK.interconnect_cml_path


def _set_active_pdk(pdk: Pdk) -> None:
    global _ACTIVE_PDK
    old_pdk = _ACTIVE_PDK
    _ACTIVE_PDK = pdk
    on_pdk_activated.fire(old_pdk=old_pdk, new_pdk=pdk)


def get_routing_strategies() -> Dict[str, Callable]:
    """Gets a dictionary of named routing functions available to the PDK, if defined, or gdsfactory defaults otherwise."""
    from gdsfactory.routing.factories import (
        routing_strategy as default_routing_strategies,
    )

    routing_strategies = get_active_pdk().routing_strategies
    if routing_strategies is None:
        routing_strategies = default_routing_strategies
    return routing_strategies


on_pdk_activated: Event = Event()
on_cell_registered: Event = Event()
on_container_registered: Event = Event()
on_yaml_cell_registered: Event = Event()
on_yaml_cell_modified: Event = Event()
on_cross_section_registered: Event = Event()

on_container_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_modified.add_handler(show)


if __name__ == "__main__":
    from gdsfactory.components import cells
    from gdsfactory.cross_section import cross_sections

    c = Pdk(
        name="demo",
        cells=cells,
        cross_sections=cross_sections,
        # layers=dict(DEVREC=(3, 0), PORTE=(3, 5)),
        sparameters_path="/home",
    )
    print(c.json())



from __future__ import annotations

import gdsfactory as gf
import gdsfactory.simulation.gtidy3d as gt
from gdsfactory.config import PATH
from gdsfactory.simulation.gtidy3d.get_results import get_results

# def test_results_run(data_regression) -> None:
#     """Run simulations and checks local results."""

#     component = gf.components.straight(length=3)
#     sim = gt.get_simulation(component=component, is_3d=False)

#     dirpath = PATH.sparameters
#     r = get_results(sim=sim, dirpath=dirpath, overwrite=True).result()

#     if data_regression:
#         data_regression.check(r.monitor_data)


if __name__ == "__main__":
    # test_results_run(None)

    component = gf.components.straight(length=3)
    sim = gt.get_simulation(component=component, is_3d=False)

    dirpath = PATH.sparameters
    r = get_results(sim=sim, dirpath=dirpath, overwrite=True).result()



try:
    from ipywidgets import (
        HTML,
        Image,
        AppLayout,
        VBox,
        HBox,
        Button,
        Label,
        Layout,
        Tab,
        Accordion,
    )
    from ipyevents import Event
    from typing import Optional
    import klayout.db as db
    import klayout.lay as lay
except ImportError as e:
    print(
        "You need install jupyter notebook plugin with `pip install gdsfactory[full]`"
    )
    raise e


class LayoutViewer:
    def __init__(
        self,
        filepath: str,
        layer_properties: Optional[str],
        hide_unused_layers: bool = True,
        with_layer_selector: bool = True,
    ):
        filepath = str(filepath)
        layer_properties = str(layer_properties)
        self.hide_unused_layers = hide_unused_layers
        self.filepath = filepath
        self.layer_properties = layer_properties

        self.layout_view = lay.LayoutView()
        self.load_layout(filepath, layer_properties)

        if self.hide_unused_layers:
            self.layout_view.remove_unused_layers()
            self.layout_view.reload_layout(self.layout_view.current_layer_list)

        pixel_buffer = self.layout_view.get_pixels_with_options(800, 600)
        png_data = pixel_buffer.to_png_data()
        self.image = Image(value=png_data, format="png")
        scroll_event = Event(source=self.image, watched_events=["wheel"])
        scroll_event.on_dom_event(self.on_scroll)
        self.wheel_info = HTML("Waiting for a scroll...")
        self.mouse_info = HTML("Waiting for a mouse event...")
        # self.layout_view.on_image_updated_event = lambda: self.refresh
        mouse_event = Event(
            source=self.image, watched_events=["mousedown", "mouseup", "mousemove"]
        )
        mouse_event.on_dom_event(self.on_mouse_down)

        if with_layer_selector:
            layer_selector_tabs = self.layer_selector_tabs = self.build_layer_selector(
                max_height=pixel_buffer.height()
            )
        else:
            layer_selector_tabs = None

        self.widget = AppLayout(
            center=self.image,
            right_sidebar=layer_selector_tabs,
            left_sidebar=None,
            # footer=VBox([self.wheel_info, self.mouse_info]),
            align_items="top",
            justify_items="left",
        )

    def button_toggle(self, button):
        button.style.button_color = (
            "transparent"
            if (button.style.button_color == button.default_color)
            else button.default_color
        )

        layer_iter = self.layout_view.begin_layers()

        while not layer_iter.at_end():
            props = layer_iter.current()
            if props.name == button.layer_props.name:
                props.visible = not props.visible
                self.layout_view.set_layer_properties(layer_iter, props)
                self.layout_view.reload_layout(self.layout_view.current_layer_list)
                break
            layer_iter.next()
        self.refresh()

    def build_layer_toggle(self, prop_iter: lay.LayerPropertiesIterator):
        from gdsfactory.utils.color_utils import ensure_six_digit_hex_color

        props = prop_iter.current()
        layer_color = ensure_six_digit_hex_color(props.eff_fill_color())

        # Would be nice to use LayoutView.icon_for_layer() rather than simple colored box
        button_layout = Layout(
            width="5px",
            height="20px",
            border=f"solid 2px {layer_color}",
            display="block",
        )

        layer_checkbox = Button(
            style={"button_color": layer_color if props.visible else "transparent"},
            layout=button_layout,
        )
        layer_checkbox.default_color = layer_color
        layer_checkbox.layer_props = props

        if props.has_children():
            prop_iter = prop_iter.down_first_child()
            n_children = prop_iter.num_siblings()
            # print(f"{props.name} has {n_children} children!")
            children = []
            for _i in range(n_children):
                prop_iter = prop_iter.next()
                children.append(self.build_layer_toggle(prop_iter))
            layer_label = Accordion([VBox(children)], titles=(props.name,))
        else:
            layer_label = Label(props.name)
        layer_checkbox.label = layer_label

        layer_checkbox.on_click(self.button_toggle)
        return HBox([layer_checkbox, layer_label])

    def build_layer_selector(self, max_height: float):
        """Builds a widget for toggling layer displays.

        Args:
            max_height: Maximum height to set for the widget (likely the height of the pixel buffer).
        """

        all_boxes = []

        prop_iter = self.layout_view.begin_layers()
        while not prop_iter.at_end():
            layer_toggle = self.build_layer_toggle(prop_iter)
            all_boxes.append(layer_toggle)
            prop_iter.next()

        layers_layout = Layout(
            max_height=f"{max_height}px", overflow_y="auto", display="block"
        )
        layer_selector = VBox(all_boxes, layout=layers_layout)

        # For when tabs are implemented
        layer_selector_tabs = Tab([layer_selector])
        layer_selector_tabs.titles = ("Layers",)
        return layer_selector_tabs

    def load_layout(self, filepath: str, layer_properties: Optional[str]):
        """Loads a GDS layout.

        Args:
            filepath: path for the GDS layout.
            layer_properties: Optional path for the layer_properties klayout file (lyp).
        """
        self.layout_view.load_layout(filepath)
        self.layout_view.max_hier()
        if layer_properties:
            self.layout_view.load_layer_props(layer_properties)

    def refresh(self):
        pixel_buffer = self.layout_view.get_pixels_with_options(800, 600)
        png_data = pixel_buffer.to_png_data()
        self.image.value = png_data

    def _get_modifier_buttons(self, event):
        shift = event["shiftKey"]
        alt = event["altKey"]
        ctrl = event["ctrlKey"]
        # meta = event["metaKey"]

        mouse_buttons = event["buttons"]

        buttons = 0
        if shift:
            buttons |= lay.ButtonState.ShiftKey
        if alt:
            buttons |= lay.ButtonState.AltKey
        if ctrl:
            buttons |= lay.ButtonState.ControlKey

        if mouse_buttons == 1:
            buttons |= lay.ButtonState.LeftButton
        elif mouse_buttons == 2:
            buttons |= lay.ButtonState.RightButton
        elif mouse_buttons == 4:
            buttons |= lay.ButtonState.MidButton

        return buttons

    def on_scroll(self, event):
        delta = event["deltaY"]
        # x = event["offsetX"]
        # y = event["offsetY"]
        self.wheel_info.value = f"scroll event: {event}"
        # buttons = self._get_modifier_buttons(event)
        # TODO: this is what I *want* to respond with, but it doesn't work, so I am using zoom_in/zoom_out instead
        # self.layout_view.send_wheel_event(delta, False, db.Point(x, y), buttons)
        if delta < 0:
            self.layout_view.zoom_in()
        else:
            self.layout_view.zoom_out()
        self.refresh()

    def on_mouse_down(self, event):
        x = event["offsetX"]
        y = event["offsetY"]
        moved_x = event["movementX"]
        moved_y = event["movementY"]
        buttons = self._get_modifier_buttons(event)
        # TODO: this part is also not working. why?
        if event == "mousedown":
            self.layout_view.send_mouse_press_event(db.Point(x, y), buttons)
        elif event == "mouseup":
            self.layout_view.send_mouse_release_event(db.Point(x, y), buttons)
        elif event == "mousemove":
            self.layout_view.send_mouse_move_event(db.Point(moved_x, moved_y), buttons)
        self.refresh()
        self.mouse_info.value = f"mouse event: {event}"



from __future__ import annotations

from gdsfactory.typings import ComponentOrPath


def get_netlist_klayout(gdspath: ComponentOrPath) -> None:
    """Returns a boolean operation between two components Uses klayout python API.

    Args:
        gdspath: path to GDS.

    """
    import klayout.db as pya

    layout = pya.Layout()
    layout.read(str(gdspath))
    return layout.top_cell()


if __name__ == "__main__":
    import klayout.db as pya

    from gdsfactory.samples.demo.lvs import pads_with_routes

    c = pads_with_routes()
    c.show()
    gdspath = c.write_gds("a.gds")

    layout = pya.Layout()
    layout.read(str(gdspath))
    c = get_netlist_klayout(gdspath)



from __future__ import annotations

import functools
from typing import Callable

import gdstk
from pydantic import validate_arguments

from gdsfactory.cell import _F, cell_without_validator
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpecs


def symbol(func: _F, *args, **kwargs) -> _F:
    """Decorator for Component symbols.

    Wraps cell_without_validator
    Validates type annotations with pydantic.

    Implements a cache so that if a symbol has already been built
    it will return the component from the cache directly.
    This avoids 2 exact cells that are not references of the same cell
    You can always over-ride this with `cache = False`.

    When decorate your functions with @cell you get:

    - CACHE: avoids creating duplicated cells.
    - name: gives Components a unique name based on parameters.
    - adds Component.info with default, changed and full component settings.

    Keyword Args:
        autoname (bool): if True renames component based on args and kwargs.
        name (str): Optional (ignored when autoname=True).
        cache (bool): returns component from the cache if it already exists.
            if False creates a new component.
            by default True avoids having duplicated cells with the same name.
        info: updates component.info dict.
        prefix: name_prefix, defaults to function name.
        max_name_length: truncates name beyond some characters (32) with a hash.
        decorator: function to run over the component.
    """
    if "prefix" not in kwargs:
        prefix = f"SYMBOL_{func.__name__}"
        kwargs["prefix"] = prefix
    _wrapped = functools.partial(
        cell_without_validator(validate_arguments(func)), **kwargs
    )
    _wrapped._symbol = True
    return _wrapped


def symbol_from_cell(func: _F, to_symbol: Callable[[Component, ...], Component]) -> _F:
    """Creates a symbol function from a component function.

    Args:
        func: the cell function
        to_symbol: the function that transforms the output of the cell function into a symbol

    Returns:
        a symbol function
    """

    @functools.wraps(func)
    def _symbol(*args, **kwargs):
        component = func(*args, **kwargs)
        symbol = to_symbol(component, prefix=f"SYMBOL_{func.__name__}")
        return symbol

    _symbol._symbol = True
    return _symbol


@symbol
def floorplan_with_block_letters(
    component: Component, copy_layers: LayerSpecs = ("WG",)
) -> Component:
    """Returns symbol with same floorplan as component layout, function name \
        and optionally shapes on layers copied from the original layout.

    Args:
        component: the layout component.
        copy_layers: if specified, copies layers from the layout into the symbol.

    Returns:
        A component representing the symbol.

    """
    import gdsfactory as gf
    from gdsfactory.components import rectangle, text

    w, h = component.size
    sym = Component()

    # add floorplan box
    bbox = sym << rectangle(size=(w, h), layer=(0, 0))
    bbox.move((0, 0), destination=component.bbox[0])

    # add text, fit to box with specified margin
    margin = 0.2
    max_w, max_h = w * (1 - margin), h * (1 - margin)
    text_init_size = 3.0
    text_init = text(
        component.settings.function_name,
        size=text_init_size,
        layer=(2, 0),
        justify="center",
    )
    w_text, h_text = text_init.size
    w_scaling = max_w / w_text
    h_scaling = max_h / h_text
    scaling = min(w_scaling, h_scaling)
    text_size = text_init_size * scaling
    text_component = text(
        component.settings.function_name, size=text_size, layer=(2, 0), justify="center"
    )

    text = sym << text_component
    text.move(text.center, destination=bbox.center)

    # add ports
    sym.add_ports(component.ports)

    # add specified layers from original layout
    if copy_layers:
        for layer in copy_layers:
            layer = gf.get_layer(layer)
            polys = component.get_polygons(by_spec=layer, as_array=False)
            # run OR to simplify shapes
            polys = gdstk.boolean(polys, [], "or", layer=layer[0], datatype=layer[1])
            sym.add_polygon(polys)

    return sym



r"""SIEPIC coupler sample.

.. code::

           n2            n4
           |             |
            \           /
             \         /
           ---=========---
        n1    length_x    n3

"""


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from simphony.libraries import siepic

    from gdsfactory.simulation.simphony import plot_model

    c = siepic.HalfRing(
        gap=200e-9, radius=12e-6, width=500e-9, thickness=220e-9, couple_length=0.0
    )
    c.rename_pins("n1", "n2", "n4", "n3")
    plot_model(c, pin_in="n1")
    print(c)
    plt.show()



from gdsfactory.config import logger
from gdsfactory.typings import (
    Callable,
    Component,
    ComponentFactory,
    ComponentFactoryDict,
    ComponentSpec,
    ComponentOrPath,
    ComponentOrReference,
    Coordinate,
    Coordinates,
    CrossSection,
    CrossSectionFactory,
    CrossSectionOrFactory,
    CrossSectionSpec,
    MultiCrossSectionAngleSpec,
    Float2,
    Float3,
    Floats,
    Int2,
    Int3,
    Ints,
    Layer,
    Label,
    Layers,
    LayerLevel,
    LayerStack,
    LayerSpec,
    LayerSpecs,
    NameToFunctionDict,
    Number,
    PathType,
    PathTypes,
    Route,
    RouteFactory,
    Routes,
    Strs,
    Section,
    Any,
    Dict,
    List,
    Optional,
    Union,
    Tuple,
)


logger.warning(
    "gdsfactory.types has been renamed to gdsfactory.typings. types will be removed in the near future.",
    DeprecationWarning,
)

__all__ = (
    "Callable",
    "Component",
    "ComponentFactory",
    "ComponentFactoryDict",
    "ComponentSpec",
    "ComponentOrPath",
    "ComponentOrReference",
    "Coordinate",
    "Coordinates",
    "CrossSection",
    "CrossSectionFactory",
    "CrossSectionOrFactory",
    "CrossSectionSpec",
    "MultiCrossSectionAngleSpec",
    "Float2",
    "Float3",
    "Floats",
    "Int2",
    "Int3",
    "Ints",
    "Layer",
    "Label",
    "Layers",
    "LayerLevel",
    "LayerStack",
    "LayerSpec",
    "LayerSpecs",
    "NameToFunctionDict",
    "Number",
    "PathType",
    "PathTypes",
    "Route",
    "RouteFactory",
    "Routes",
    "Strs",
    "Section",
    "Any",
    "Dict",
    "List",
    "Optional",
    "Union",
    "Tuple",
)



"""Straight waveguide."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import CrossSectionSpec


@gf.cell
def straight(
    length: float = 10.0,
    npoints: int = 2,
    with_bbox: bool = True,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    """Returns a Straight waveguide.

    Args:
        length: straight length (um).
        npoints: number of points.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: specification (CrossSection, string, CrossSectionFactory dict).
        kwargs: cross_section settings.

    .. code::

        o1 -------------- o2
                length
    """
    length = snap_to_grid(length)
    p = gf.path.straight(length=length, npoints=npoints)
    x = gf.get_cross_section(cross_section, **kwargs)

    c = Component()
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)
    c.info["length"] = length
    c.info["width"] = x.width
    c.info["cross_section"] = cross_section

    if x.info:
        c.info.update(x.info)

    if with_bbox and length and x.bbox_layers:
        padding = []
        for offset in x.bbox_offsets:
            points = get_padding_points(
                component=c,
                default=0,
                bottom=offset,
                top=offset,
            )
            padding.append(points)

        for layer, points in zip(x.bbox_layers, padding):
            c.add_polygon(points, layer=layer)
    c.absorb(ref)
    return c


if __name__ == "__main__":
    # c = straight(cross_section=gf.partial('metal_routing', width=2), with_hash=False)
    # c = straight(
    #     cross_section=gf.partial(gf.cross_section.strip, width=2), with_hash=False
    # )
    # c = straight(cladding_offset=2.5)

    nm = 1e-3
    c = straight(width=101 * nm)
    print(c.name)

    # strip2 = gf.get_cross_section("strip", layer=(2, 0))
    # settings = dict(width=2)

    # c = straight(
    #     length=1, cross_section={"cross_section": "strip", "settings": settings}
    # )
    # c = straight(
    #     length=1,
    #     cross_section={"cross_section": "strip", "settings": settings},
    #     width=3,
    #     # bbox_layers=[(2, 0)],
    #     # bbox_offsets=[3],
    # )
    # c.assert_ports_on_grid()
    c.show(show_ports=True)
    # c.plot()
    # c.pprint()



"""Preprocessing involving mostly the GDS polygons."""
from __future__ import annotations

import gdsfactory as gf
import shapely
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Polygon


def round_coordinates(geom, ndigits=4):
    """Round coordinates to n_digits to eliminate floating point errors."""

    def _round_coords(x, y, z=None):
        x = round(x, ndigits)
        y = round(y, ndigits)

        if z is not None:
            z = round(x, ndigits)

        return [c for c in (x, y, z) if c is not None]

    return shapely.ops.transform(_round_coords, geom)


def fuse_polygons(
    component, layername, layer, round_tol=4, simplify_tol=1e-4, offset_tol=None
):
    """Take all polygons from a layer, and returns a single (Multi)Polygon shapely object."""
    layer_component = component.extract(layer)

    # gdstk union before shapely conversion helps with ill-formed polygons
    offset_tol = offset_tol or gf.get_active_pdk().grid_size
    layer_component = gf.geometry.offset(
        layer_component, distance=offset_tol, precision=1e-6, layer=layer
    )
    layer_component = gf.geometry.offset(
        layer_component, distance=-offset_tol, precision=1e-6, layer=layer
    )

    shapely_polygons = [
        round_coordinates(shapely.geometry.Polygon(polygon), round_tol)
        for polygon in layer_component.get_polygons()
    ]

    return shapely.ops.unary_union(shapely_polygons).simplify(
        simplify_tol, preserve_topology=True
    )


def cleanup_component(component, layerstack, round_tol=2, simplify_tol=1e-2):
    """Process component polygons before meshing."""
    layerstack_dict = layerstack.to_dict()
    return {
        layername: fuse_polygons(
            component,
            layername,
            layer["layer"],
            round_tol=round_tol,
            simplify_tol=simplify_tol,
        )
        for layername, layer in layerstack_dict.items()
        if layer["layer"] is not None
    }


def to_polygons(geometries):
    for geometry in geometries:
        if isinstance(geometry, Polygon):
            yield geometry
        else:
            yield from geometry.geoms


def to_lines(geometries):
    for geometry in geometries:
        if isinstance(geometry, LineString):
            yield geometry
        else:
            yield from geometry.geoms


def tile_shapes(shapes_dict):
    """Break up shapes in order so that plane is tiled with non-overlapping layers."""
    shapes_tiled_dict = {}
    for lower_index, (lower_name, lower_shapes) in reversed(
        list(enumerate(shapes_dict.items()))
    ):
        tiled_lower_shapes = []
        for lower_shape in (
            lower_shapes.geoms if hasattr(lower_shapes, "geoms") else [lower_shapes]
        ):
            diff_shape = lower_shape
            for _higher_index, (_higher_name, higher_shapes) in reversed(
                list(enumerate(shapes_dict.items()))[:lower_index]
            ):
                for higher_shape in (
                    higher_shapes.geoms
                    if hasattr(higher_shapes, "geoms")
                    else [higher_shapes]
                ):
                    diff_shape = diff_shape.difference(higher_shape)
            tiled_lower_shapes.append(diff_shape)
        if tiled_lower_shapes and tiled_lower_shapes[0].geom_type in [
            "Polygon",
            "MultiPolygon",
        ]:
            shapes_tiled_dict[lower_name] = MultiPolygon(
                to_polygons(tiled_lower_shapes)
            )
        else:
            shapes_tiled_dict[lower_name] = MultiLineString(tiled_lower_shapes)

    return shapes_tiled_dict



"""
Reference: Selberherr, S. (1984). Process Modeling. In: Analysis and Simulation of Semiconductor Devices. Springer, Vienna. https://doi.org/10.1007/978-3-7091-8752-4_3
"""
import numpy as np

from gdsfactory.simulation.process.silicon import ni


diffusion_in_silicon = {
    "boron": {
        "D00": 0.037,  # cm2s-1
        "Ea0": 3.46,  # eV
        "D0-": 0.0,
        "Ea-": 0.0,
        "D0=": 0.0,
        "Ea=": 0.0,
        "D0+": 0.72,
        "Ea+": 3.46,
    },
    "phosphorus": {
        "D00": 3.85,  # cm2s-1
        "Ea0": 3.66,  # eV
        "D0-": 4.44,
        "Ea-": 4.0,
        "D0=": 44.20,
        "Ea=": 4.37,
        "D0+": 0.0,
        "Ea+": 0.0,
    },
    "antimony": {
        "D00": 0.214,  # cm2s-1
        "Ea0": 3.65,  # eV
        "D0-": 15.0,
        "Ea-": 4.08,
        "D0=": 0.0,
        "Ea=": 0.0,
        "D0+": 0.0,
        "Ea+": 0.0,
    },
    "arsenic": {
        "D00": 0.066,  # cm2s-1
        "Ea0": 3.44,  # eV
        "D0-": 12.0,
        "Ea-": 4.05,
        "D0=": 0.0,
        "Ea=": 0.0,
        "D0+": 0.0,
        "Ea+": 0.0,
    },
}


def D(
    dopant: str,
    T: float,
    n: float = None,
    p: float = None,
):
    """
    Diffusion coefficient of dopants in silicon.

    Arguments:
        dopant: dopant atom name
        T: temperature (C)
        n: donor concentration (/cm3), defaults to ni
        p: acceptor concentration (/cm3), defaults to ni
    """
    kB = 8.617333262e-5  # eV K-1
    T_kelvin = T + 273.15
    if n is None:
        n = ni(T_kelvin)
    if p is None:
        p = ni(T_kelvin)
    Di0 = diffusion_in_silicon[dopant]["D00"] * np.exp(
        -1 * diffusion_in_silicon[dopant]["Ea0"] / (kB * T_kelvin)
    )
    Din = diffusion_in_silicon[dopant]["D0-"] * np.exp(
        -1 * diffusion_in_silicon[dopant]["Ea-"] / (kB * T_kelvin)
    )
    Dinn = diffusion_in_silicon[dopant]["D0="] * np.exp(
        -1 * diffusion_in_silicon[dopant]["Ea="] / (kB * T_kelvin)
    )
    Dip = diffusion_in_silicon[dopant]["D0+"] * np.exp(
        -1 * diffusion_in_silicon[dopant]["Ea+"] / (kB * T_kelvin)
    )
    return (
        Di0
        + Din * n / ni(T_kelvin)
        + Dinn * (n / ni(T_kelvin)) ** 2
        + Dip * p / ni(T_kelvin)
    )


def silicon_diffused_gaussian_profile(
    dopant: str,
    dose: float,
    E: float,
    t: float,
    T: float,
    z: np.array = np.linspace(0, 1, 1000),
    # x: np.array = np.linspace(-5,5,1000),
):
    """
    Returns diffused gaussian implantation profile for dopant in silicon.

    Arguments:
        dopant: str name of implant
        dose: implant dose per unit area (/cm2)
        E: energy of implant (keV)
        t: diffusion time (s)
        T: temperature applied during diffusion time (C)
        z: depth coordinate (um)

    Returns:
        C(z): ion distribution as a function of substrate position and depth (ions/cm3)
    """
    from gdsfactory.simulation.process.implant_tables import (
        depth_in_silicon,
        straggle_in_silicon,
    )

    z0 = depth_in_silicon[dopant](E)
    dz = straggle_in_silicon[dopant](E)
    Dconst = D(dopant=dopant, T=T) * 1e8  # convert from cm2/s to to um2/s

    # 2D functions (TODO)
    # a = np.mean(x)
    # xx, zz = np.meshgrid(x, z)
    # def H(z,t):
    # return np.exp(-(z - z0)**2 / (2 * dz**2 + 4*Dconst*t)) * special.erf(-z0/dz * np.sqrt(Dconst*t/(dz**2 + 2*Dconst*t)) - np.sqrt(dz**2 / (4*Dconst*t*(dz**2 + 2*Dconst*t))) * z)
    # return dose / (4 * np.sqrt(2 * np.pi * (dz**2 + 2*Dconst*t))) * (H(z,t) + H(-1*z,t)) * special.erf((a - xx)/(2 * np.sqrt(Dconst * t)))

    return (
        dose
        / (np.sqrt(2 * np.pi * (dz**2 + 2 * Dconst * t)))
        * np.exp(-((z - z0) ** 2) / (dz**2 + 2 * Dconst * t))
    )


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Ts = [800,850,900,950,1000,1050]
    # conc = np.logspace(18,21,100)

    # for T in Ts:
    #     plt.loglog(conc, D("phosphorus", T, n=conc, p=conc), label=T)

    # plt.xlabel("Acceptor concentration (cm-3)")
    # plt.ylabel("Diffusivity (cm2 s-1)")
    # plt.title("Intrinsic diffusivity (n=p=ni)")
    # plt.legend()
    # plt.show()

    for t in [0, 60, 5 * 60, 10 * 60]:
        conc = silicon_diffused_gaussian_profile(
            dopant="phosphorus",
            dose=1e12,
            E=100,
            t=t,
            T=1000,
            z=np.linspace(0, 0.6, 1000),
        )
        plt.plot(conc, label="t")
    plt.title("Phosphorus 1E12 ions/cm2, 100keV, 1000C anneal")
    plt.xlabel("depth (um)")
    plt.ylabel("ion concentration (cm-3)")
    plt.legend(title="Diffusion time (s)")
    plt.show()



"""Bends with grating couplers inside the spiral.

maybe: need to add grating coupler loopback as well
"""
from __future__ import annotations

from typing import Optional, Tuple

import numpy as np
from numpy import float64

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


def get_bend_port_distances(bend: Component) -> Tuple[float64, float64]:
    """Returns distance between bend ports."""
    p0, p1 = bend.ports.values()
    return abs(p0.x - p1.x), abs(p0.y - p1.y)


@cell
def spiral_external_io(
    N: int = 6,
    x_inner_length_cutback: float = 300.0,
    x_inner_offset: float = 0.0,
    y_straight_inner_top: float = 0.0,
    xspacing: float = 3.0,
    yspacing: float = 3.0,
    bend: ComponentSpec = bend_euler,
    length: Optional[float] = None,
    cross_section: CrossSectionSpec = "strip",
    with_inner_ports: bool = False,
    y_straight_outer_offset: float = 0.0,
    inner_loop_spacing_offset: float = 0.0,
    **kwargs,
) -> Component:
    """Returns spiral with input and output ports outside the spiral.

    Args:
        N: number of loops.
        x_inner_length_cutback: x inner length.
        x_inner_offset: x inner offset.
        y_straight_inner_top: y straight inner top.
        xspacing: center to center x-spacing.
        yspacing: center to center y-spacing.
        bend: function.
        length: length in um, it is the approximates total length.
        cross_section: spec.
        with_inner_ports: if True, removes the internal S-bend and exposes new ports
        y_straight_outer_offset: amount to add/remove to the last points at the outer output of the spiral
        inner_loop_spacing_offset: extra difference between the inner ports
        kwargs: cross_section settings.
    """
    if length:
        x_inner_length_cutback = length / (4 * (N - 1))

    y_straight_inner_top += 5
    x_inner_length_cutback += x_inner_offset
    _bend180 = bend(angle=180, cross_section=cross_section, **kwargs)
    _bend90 = bend(angle=90, cross_section=cross_section, **kwargs)

    bend_radius = _bend90.info["radius"]
    rx, ry = get_bend_port_distances(_bend90)
    _, rx180 = get_bend_port_distances(_bend180)  # rx180, second arg since we rotate

    component = Component()
    inner_loop_spacing = 2 * bend_radius + 5.0 + inner_loop_spacing_offset

    # Create manhattan path going from west grating to westmost port of bend 180
    x_inner_length = x_inner_length_cutback + 5.0 + xspacing

    y_inner_bend = y_straight_inner_top - bend_radius - 5.0
    x_inner_loop = x_inner_length - 5.0
    p1 = (x_inner_loop, y_inner_bend)
    p2 = (x_inner_loop + inner_loop_spacing, y_inner_bend)

    _pt = np.array(p1)
    pts_w = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i + 1) * yspacing
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + (2 * i + 1) * xspacing
        y3 = -ry - (2 * i + 2) * yspacing
        y3 += 0 if i < N - 1 else y_straight_outer_offset
        x4 = -(2 * i + 1) * xspacing
        if i == N - 1:
            x4 = x4 - rx180 + xspacing

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_w += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_w = pts_w[:-2]

    # Create manhattan path going from east grating to eastmost port of bend 180
    _pt = np.array(p2)
    pts_e = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i) * yspacing
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + 2 * i * xspacing
        y3 = -ry - (2 * i + 1) * yspacing
        y3 += 0 if i < N - 1 else y_straight_outer_offset
        x4 = -2 * i * xspacing

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_e += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_e = pts_e[:-2]

    # Join the two bits of paths and extrude the spiral geometry
    if not with_inner_ports:
        route = round_corners(
            pts_w[::-1] + pts_e,
            bend=bend,
            cross_section=cross_section,
            **kwargs,
        )
        component.add(route.references)
        component.add_port("o2", port=route.ports[0])
        component.add_port("o1", port=route.ports[1])

        length = route.length
    # If inner ports, do not join and layout routes separately
    else:
        pts_w[1][0] += bend_radius
        pts_e[1][0] += bend_radius
        route_west = round_corners(
            pts_w[1:],
            bend=bend,
            cross_section=cross_section,
            **kwargs,
        )
        route_east = round_corners(
            pts_e[1:],
            bend=bend,
            cross_section=cross_section,
            **kwargs,
        )
        component.add(route_west.references)
        component.add(route_east.references)
        component.add_port("o2", port=route_west.ports[0])
        component.add_port("o1", port=route_east.ports[1])
        component.add_port("o4", port=route_west.ports[1])
        component.add_port("o3", port=route_east.ports[0])
        length = route_west.length + route_east.length

    component.info["length"] = length

    return component


if __name__ == "__main__":
    spacing = 3
    c = spiral_external_io(
        N=15,
        xspacing=spacing,
        yspacing=spacing,
        with_inner_ports=True,
        x_inner_length_cutback=0,
        y_straight_inner_top=0,
        x_inner_offset=0,
    )
    c.show(show_ports=True)



from __future__ import annotations

import re
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import logger
from gdsfactory.generic_tech import LAYER_STACK
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_lumerical as get_sparameters_path,
)
from gdsfactory.technology import LayerStack


def get_ports(line: str) -> Tuple[str, str]:
    """Returns 2 port labels strings from interconnect file."""
    line = line.replace('"', "")
    line = line.replace("(", "")
    line_fields = line.split(",")
    port1 = line_fields[0]
    port2 = line_fields[3]
    return port1, port2


def read_sparameters_file(
    filepath, numports: int
) -> Tuple[Tuple[str, ...], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect export file.

    Args:
        filepath: Sparameters filepath (interconnect format).
        numports: number of ports.

    Returns:
        port_names: list of port labels.
        F: frequency 1d np.array.
        S: Sparameters np.ndarray matrix.

    """
    F = []
    S = []
    port_names = []

    with open(filepath) as fid:
        for _i in range(numports):
            port_line = fid.readline()
            m = re.search(r'\[".*",', port_line)
            if m:
                port = m[0]
                port_names.append(port[2:-2])
        line = fid.readline()
        port1, port2 = get_ports(line)
        line = fid.readline()
        numrows = int(tuple(line[1:-2].split(","))[0])
        S = np.zeros((numrows, numports, numports), dtype="complex128")
        r = m = n = 0
        for line in fid:
            if line[0] == "(":
                if "transmission" in line:
                    port1, port2 = get_ports(line)
                continue
            data = line.split()
            data = list(map(float, data))
            if m == 0 and n == 0:
                F.append(data[0])

            i = port_names.index(port1)
            j = port_names.index(port2)
            S[r, i, j] = data[1] * np.exp(1j * data[2])
            r += 1
            if r == numrows:
                r = 0
                m += 1
                if m == numports:
                    m = 0
                    n += 1
                    if n == numports:
                        break

    # port_names.reverse()
    # print(len(F), S.shape, len(port_names))
    return tuple(port_names), np.array(F), S


def read_sparameters_lumerical(
    component: Optional[Component] = None,
    layer_stack: LayerStack = LAYER_STACK,
    filepath: Optional[str] = None,
    numports: Optional[int] = None,
    dirpath: Path = gf.PATH.sparameters,
    **kwargs,
) -> Tuple[List[str], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect .DAT file.

    Args:
        component: Component.
        layer_stack: layer thickness and material.
        filepath: for file.
        numports: number of ports.
        dirpath: path where to look for the Sparameters.

    Keyword Args:
        simulation_settings.

    Returns:
        port_names: list of port labels.
        F: frequency 1d np.array.
        S: Sparameters np.ndarray matrix.


    the Sparameters file have Lumerical format
    https://support.lumerical.com/hc/en-us/articles/360036107914-Optical-N-Port-S-Parameter-SPAR-INTERCONNECT-Element#toc_5
    """
    if component is None and filepath is None:
        raise ValueError("You need to define the filepath or the component")

    if filepath and numports is None:
        raise ValueError("You need to define numports")

    filepath = filepath or get_sparameters_path(
        component=component, dirpath=dirpath, layer_stack=layer_stack, **kwargs
    ).with_suffix(".dat")
    numports = numports or len(component.ports)
    if not filepath.exists():
        raise ValueError(f"Sparameters for {component.name!r} not found in {filepath}")
    assert numports > 1, f"number of ports = {numports} and needs to be > 1"

    logger.info(f"Sparameters loaded from {filepath}")
    return read_sparameters_file(filepath=filepath, numports=numports)


if __name__ == "__main__":
    r = read_sparameters_lumerical(gf.components.mmi1x2())



from __future__ import annotations

from typing import Awaitable
import time

import numpy as np
import tidy3d as td
from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.config import logger
from gdsfactory.serialization import clean_value_json
from gdsfactory.simulation import port_symmetries
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_tidy3d as get_sparameters_path,
)
from gdsfactory.simulation.gtidy3d.get_results import _executor, get_results_batch
from gdsfactory.simulation.gtidy3d.get_simulation import get_simulation, plot_simulation
from gdsfactory.typings import (
    Any,
    ComponentSpec,
    Dict,
    List,
    Optional,
    PathType,
    Port,
    PortSymmetries,
    Tuple,
    Sparameters,
)


def parse_port_eigenmode_coeff(
    port_name: str, ports: Dict[str, Port], sim_data: td.SimulationData
) -> Tuple[np.ndarray]:
    """Given a port and eigenmode coefficient result, returns the coefficients \
    relative to whether the wavevector is entering or exiting simulation.

    Args:
        port_name: port name.
        ports: component_ref.ports.
        sim_data: simulation data.
    """
    # Direction of port (pointing away from the simulation)
    # Figure out if that is exiting the simulation or not
    # depending on the port orientation (assuming it's near PMLs)

    orientation = ports[port_name].orientation

    if orientation in [0, 90]:  # east
        direction_inp = "-"
        direction_out = "+"
    elif orientation in [180, 270]:  # west
        direction_inp = "+"
        direction_out = "-"
    else:
        raise ValueError(
            "Port orientation = {orientation} is not 0, 90, 180, or 270 degrees"
        )

    coeff_inp = sim_data.monitor_data[port_name].amps.sel(direction=direction_inp)
    coeff_out = sim_data.monitor_data[port_name].amps.sel(direction=direction_out)
    return coeff_inp.values.flatten(), coeff_out.values.flatten()


def get_wavelengths(port_name: str, sim_data: td.SimulationData) -> np.ndarray:
    coeff_inp = sim_data.monitor_data[port_name].amps.sel(direction="+")
    freqs = coeff_inp.f
    return td.constants.C_0 / freqs.values


def write_sparameters(
    component: ComponentSpec,
    port_symmetries: Optional[PortSymmetries] = None,
    port_source_names: Optional[List[str]] = None,
    dirpath: Optional[PathType] = None,
    run: bool = True,
    overwrite: bool = False,
    verbose: bool = False,
    **kwargs,
) -> Sparameters:
    """Get full sparameter matrix from a gdsfactory Component.

    Simulates each time using a different input port (by default, all of them)
    unless you specify port_symmetries.

    port_symmetries = {"o1":
            {
                "s11": ["s22","s33","s44"],
                "s21": ["s21","s34","s43"],
                "s31": ["s13","s24","s42"],
                "s41": ["s14","s23","s32"],
            }
        }
    - Only simulations using the outer key port names will be run
    - The associated value is another dict whose keys are the S-parameters computed
        when this source is active
    - The values of this inner Dict are lists of s-parameters whose values are copied

    Args:
        component: to simulate.
        port_source_names: list of ports to excite. Defaults to all.
        port_symmetries: Dict to specify port symmetries, to save number of simulations
        dirpath: directory to store sparameters in npz.
            Defaults to active Pdk.sparameters_path.
        run: runs simulation, if False, only plots simulation.
        overwrite: overwrites stored Sparameter npz results.
        verbose: prints info messages and progressbars.

    Keyword Args:
        port_extension: extend ports beyond the PML.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        thickness_pml: PML thickness (um).
        xmargin: left/right distance from component to PML.
        xmargin_left: left distance from component to PML.
        xmargin_right: right distance from component to PML.
        ymargin: left/right distance from component to PML.
        ymargin_top: top distance from component to PML.
        ymargin_bot: bottom distance from component to PML.
        zmargin: thickness for cladding above and below core.
        clad_material: material for cladding.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before monitors.
        wavelength_start: in (um).
        wavelength_stop: in (um).
        wavelength_points: in (um).
        plot_modes: plot source modes.
        num_modes: number of modes to plot.
        run_time_ps: make sure it's sufficient for the fields to decay.
            defaults to 10ps and counts on automatic shutoff to stop earlier if needed.
        dispersive: False uses constant refractive index materials.
            True adds wavelength depending materials.
            Dispersive materials require more computation.
        material_name_to_tidy3d_index: not dispersive materials have a constant index.
        material_name_to_tidy3d_name: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with tidy3d material database names.
        is_3d: if False, does not consider Z dimension for faster simulations.
        with_all_monitors: True adds field monitor which increases results file size.
        grid_spec: defaults to automatic td.GridSpec.auto(wavelength=wavelength)
            td.GridSpec.uniform(dl=20*nm)
            td.GridSpec(
                grid_x = td.UniformGrid(dl=0.04),
                grid_y = td.AutoGrid(min_steps_per_wvl=20),
                grid_z = td.AutoGrid(min_steps_per_wvl=20),
                wavelength=wavelength,
                override_structures=[refine_box]
            )
        dilation: float = 0.0
            Dilation of the polygon in the base by shifting each edge along its
            normal outwards direction by a distance;
            a negative value corresponds to erosion.
        sidewall_angle_deg : float = 0
            Angle of the sidewall.
            ``sidewall_angle=0`` (default) specifies vertical wall,
            while ``0<sidewall_angle_deg<90`` for the base to be larger than the top.

    """
    component = gf.get_component(component)
    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        **kwargs,
    )
    filepath_sim_settings = filepath.with_suffix(".yml")
    if filepath.exists() and not overwrite and run:
        logger.info(f"Simulation loaded from {filepath!r}")
        return dict(np.load(filepath))

    port_symmetries = port_symmetries or {}
    component_ref = component.ref()
    ports = component_ref.ports
    port_names = [port.name for port in list(ports.values())]

    sims = []
    sp = {}

    port_source_names = port_source_names or port_names

    for port_name in port_source_names:
        if port_name not in port_symmetries:
            sim = get_simulation(component, port_source_name=port_name, **kwargs)
            sims.append(sim)

    if not run:
        sim = sims[0]
        plot_simulation(sim)
        return sp

    start = time.time()
    batch_data = get_results_batch(sims, verbose=verbose)

    def get_sparameter(
        port_name_source: str,
        sim_data: td.SimulationData,
        port_symmetries=port_symmetries,
        **kwargs,
    ) -> Dict[str, np.ndarray]:
        """Return Component sparameter for a particular port Index n.

        Args:
            port_name: source port name.
            sim_data: simulation data.
            port_symmetries: to save simulations.
            kwargs: simulation settings.
        """
        source_entering, source_exiting = parse_port_eigenmode_coeff(
            port_name=port_name_source, ports=component_ref.ports, sim_data=sim_data
        )

        for port_name in port_names:
            monitor_entering, monitor_exiting = parse_port_eigenmode_coeff(
                port_name=port_name, ports=ports, sim_data=sim_data
            )
            sij = monitor_exiting / source_entering
            key = f"{port_name}@0,{port_name_source}@0"
            sp[key] = sij
            sp["wavelengths"] = get_wavelengths(port_name=port_name, sim_data=sim_data)

        if bool(port_symmetries):
            for key, symmetries in port_symmetries.items():
                for sym in symmetries:
                    if key in sp:
                        sp[sym] = sp[key]

        return sp

    for port_source_name, (_sim_name, sim_data) in zip(
        port_source_names, batch_data.items()
    ):
        sp.update(get_sparameter(port_source_name, sim_data))

    end = time.time()
    np.savez_compressed(filepath, **sp)
    kwargs.update(compute_time_seconds=end - start)
    kwargs.update(compute_time_minutes=(end - start) / 60)
    filepath_sim_settings.write_text(OmegaConf.to_yaml(clean_value_json(kwargs)))
    logger.info(f"Write simulation results to {str(filepath)!r}")
    logger.info(f"Write simulation settings to {str(filepath_sim_settings)!r}")
    return sp


def write_sparameters_batch(
    jobs: List[Dict[str, Any]], **kwargs
) -> List[Awaitable[Sparameters]]:
    """Returns Sparameters for a list of write_sparameters.

    Each job runs in separate thread and is non blocking.
    You need to get the results using sp.result().

    Args:
        jobs: list of kwargs for write_sparameters_grating_coupler.
        kwargs: simulation settings.
    """
    kwargs.update(verbose=False)
    return [_executor.submit(write_sparameters, **job, **kwargs) for job in jobs]


write_sparameters_1x1 = gf.partial(
    write_sparameters, port_symmetries=port_symmetries.port_symmetries_1x1
)
write_sparameters_crossing = gf.partial(
    write_sparameters, port_symmetries=port_symmetries.port_symmetries_crossing
)

write_sparameters_batch_1x1 = gf.partial(
    write_sparameters_batch, port_symmetries=port_symmetries.port_symmetries_1x1
)


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.simulation as sim

    # c = gf.components.straight(length=2.1)
    c = gf.c.straight()
    c = gf.components.mmi1x2()
    sp = write_sparameters(c, is_3d=True, port_source_names=None, overwrite=False)
    sim.plot.plot_sparameters(sp)

    # t = sp.o1@0,o2@0
    # print(f"Transmission = {t}")
    # cs = [gf.c.straight(length=1.11 + i) for i in [1, 2]]
    # sps = write_sparameters_batch_1x1(cs)



from __future__ import annotations

import warnings
from typing import Optional

from numpy import floor

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via import via1, viac
from gdsfactory.typings import ComponentSpec, Floats, LayerSpec, LayerSpecs


@gf.cell
def via_stack_slot(
    size=(11.0, 11.0),
    layers: LayerSpecs = ("M1", "M2"),
    layer_offsets: Optional[Floats] = (0, 1.0),
    layer_offsetsx: Optional[Floats] = None,
    layer_offsetsy: Optional[Floats] = None,
    layer_port: LayerSpec = None,
    via: ComponentSpec = via1,
    enclosure: float = 1.0,
    ysize: float = 0.5,
    yspacing: float = 2.0,
) -> Component:
    """Rectangular via_stack with slotted via in X direction.

    Args:
        size: of the layers.
        layers: layers on which to draw rectangles.
        layer_offsets: cladding_offset for each layer.
        layer_offsetsx: optional xoffset for layers, defaults to layer_offsets.
        layer_offsetsy: optional yoffset for layers, defaults to layer_offsets.
        layer_port: if None assumes port is on the last layer.
        via: via to use to fill the rectangles.
        enclosure: of the via by rectangle.
        ysize: via height in y.
        yspacing: via spacing pitch in y.

    .. code::

         __________________________________________
        |                |                        |
        |                | layer_offsetsy[1]      |
        |  ______________|______________________  |
        |  |<--->                              |<>|
        |  |enclosure                          | layer_offsetsx[1]
        |  |      ______________________       |  |
        |  |     |                      |      |  |
        |  |     |     via              | ysize|  |
        |  |     |______________________|      |  |
        |  |  |                                |  |
        |  |  | yspacing                size[1]|  |
        |  |  |                                |  |
        |  |  |   ______________________       |  |
        |  |  |  |                      |      |  |
        |  |  |  |     via              | ysize|  |
        |  |  |  |______________________|      |  |
        |  |                                   |  |
        |  |                                   |  |
        |  |___________________________________|  |
        |                  size[0]                |
        |                                         |
        |_________________________________________|

    """
    if size[0] - 2 * enclosure < 0:
        raise ValueError(
            f"via_stack length (size[0] = {size[0]}) < 2*enclosure ({2*enclosure}). "
        )
    if size[1] - 2 * enclosure < 0:
        raise ValueError(
            f"via_stack width (size[1] = {size[1]}) < 2*enclosure ({2*enclosure}). "
        )

    layer_port = layer_port or layers[-1]

    c = Component()
    c.info["size"] = (float(size[0]), float(size[1]))

    layer_offsetsx = layer_offsetsx or layer_offsets
    layer_offsetsy = layer_offsetsy or layer_offsets

    layer_offsetsx = list(layer_offsetsx) + [0] * len(layers)
    layer_offsetsy = list(layer_offsetsy) + [0] * len(layers)

    elements = {
        len(layers),
        len(layer_offsets),
        len(layer_offsetsx),
        len(layer_offsetsy),
    }
    if len(elements) > 1:
        warnings.warn(
            f"Got {len(layers)} layers, {len(layer_offsets)} layer_offsets, {len(layer_offsetsx)} layer_offsetsx, {len(layer_offsetsy)} layer_offsetsy."
        )

    for layer, offsetx, offsety in zip(layers, layer_offsetsx, layer_offsetsy):
        ref = c << compass(
            size=(size[0] + 2 * offsetx, size[1] + 2 * offsety), layer=layer
        )

        if layer == layer_port:
            c.add_ports(ref.ports)

    via = gf.get_component(via, size=(size[0] - 2 * enclosure, ysize))

    nb_vias_y = (size[1] - 2 * enclosure) / yspacing
    nb_vias_y = int(floor(nb_vias_y)) or 1
    ref = c.add_array(via, columns=1, rows=nb_vias_y, spacing=(0, yspacing))
    dy = (size[1] - (nb_vias_y - 1) * yspacing - size[1]) / 2
    ref.move((0, dy))
    return c


via_stack_slot_m1_m2 = gf.partial(via_stack_slot, layers=("M1", "M2"), via=via1)

via_stack_slot_slab_m1 = gf.partial(via_stack_slot, layers=("M1",), via=viac)


if __name__ == "__main__":
    # c = via_stack_slot_m1_m2(layer_offsets=(0.5, 1), enclosure=1, size=(3, 3))
    # c = via_stack_slot_m1_m2()
    c = via_stack_slot_slab_m1()
    c.show(show_ports=True)



from __future__ import annotations

from functools import partial
from typing import Dict, Optional, Union

import meep as mp
import meep.materials as mat
import numpy as np

from gdsfactory.materials import material_name_to_meep as material_name_to_meep_default

MATERIALS = [m for m in dir(mat) if not m.startswith("_")]


def get_material(
    name: str = "si",
    wavelength: float = 1.55,
    dispersive: bool = False,
    material_name_to_meep: Optional[Dict[str, Union[str, float]]] = None,
) -> mp.Medium:
    """Returns Meep Medium from database.

    Args:
        name: material name.
        wavelength: wavelength (um).
        dispersive: True for built-in Meep index model,
            False for simple, non-dispersive model.
        material_name_to_meep: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with meep material database names.

    Note:
        Using the built-in models can be problematic at low resolution.

    """
    material_name_to_meep_new = material_name_to_meep or {}
    material_name_to_meep = material_name_to_meep_default.copy()
    material_name_to_meep.update(**material_name_to_meep_new)

    materials = [material.lower() for material in MATERIALS]
    name = name.lower()

    if name not in material_name_to_meep and name not in materials:
        raise KeyError(f"material {name!r} not found in available materials")

    meep_name = material_name_to_meep[name]

    if isinstance(meep_name, (int, float)):
        # if material is only a number, we can return early regardless of dispersion
        return mp.Medium(index=meep_name)

    material = getattr(mat, meep_name)

    if dispersive:
        return material

    # now what's left is the case of having a dispersive meep medium but a simulation
    # without dispersion, so we extract the permittivity at the correct wavelength
    try:
        return mp.Medium(epsilon=material.epsilon(1 / wavelength)[0][0])
    except ValueError as e:
        print(f"material = {name!r} not defined for wavelength={wavelength}")
        raise e


def get_index(
    wavelength: float = 1.55,
    name: str = "Si",
    dispersive: bool = False,
) -> float:
    """Returns refractive index from Meep's material database.

    Args:
        name: material name.
        wavelength: wavelength (um).
        dispersive: True for built-in Meep index model,
            False for simple, non-dispersive model.

    Note:
        Using the built-in models can be problematic at low resolution.
        If fields are NaN or Inf, increase resolution or use a non-dispersive model.

    """
    medium = get_material(
        name=name,
        wavelength=wavelength,
        dispersive=dispersive,
    )

    epsilon_matrix = medium.epsilon(1 / wavelength)
    epsilon11 = epsilon_matrix[0][0]
    return float(epsilon11.real**0.5)


def test_index() -> None:
    n = get_index(name="sin")
    n_reference = 1.9962797317138816
    assert np.isclose(n, n_reference), n


si = partial(get_index, name="Si")
sio2 = partial(get_index, name="SiO2")


if __name__ == "__main__":
    test_index()
    # n = get_index(name="Si", wavelength=1.31)
    # print(n, type(n))



try:
    import kfactory as kf
    from kfactory import kdb
except ImportError as e:
    print(
        "You can install `pip install gdsfactory[full]` for using maskprep. "
        "And make sure you use python >= 3.10"
    )
    raise e

from gdsfactory.typings import Dict, Layer, PathType, Tuple, LayerSpecs


def size(region: kdb.Region, offset: float):
    return region.dup().size(int(offset * 1e3))


def boolean_or(region1: kdb.Region, region2: kdb.Region):
    return region1.__or__(region2)


def boolean_not(region1: kdb.Region, region2: kdb.Region):
    return kdb.Region.__sub__(region1, region2)


def copy(region: kdb.Region):
    return region.dup()


class Region(kdb.Region):
    def __iadd__(self, offset):
        """Adds an offset to the layer."""
        return size(self, offset)

    def __isub__(self, offset):
        """Adds an offset to the layer."""
        return size(self, offset)

    def __add__(self, element):
        if isinstance(element, (float, int)):
            return size(self, element)

        elif isinstance(element, kdb.Region):
            return boolean_or(self, element)
        else:
            raise ValueError(f"Cannot add type {type(element)} to region")

    def __sub__(self, element):
        if isinstance(element, (float, int)):
            return size(self, -element)

        elif isinstance(element, kdb.Region):
            return boolean_not(self, element)

    def copy(self):
        return self.dup()


class RegionCollection:
    """RegionCollection of region per layer.

    Args:
        layermap: dict of layernames to layer numbers {'WG': (1, 0)}
        filepath: to read GDS from.
    """

    def __init__(self, layermap: Dict[str, Layer], filepath: PathType = None):
        lib = kf.kcell.KLib()
        lib.read(filename=str(filepath))
        c = lib[0]

        for layername, layer in layermap.items():
            region = Region()
            layer = lib.layer(layer[0], layer[1])
            region.insert(c.begin_shapes_rec(layer))
            region.merge()
            setattr(self, layername, region)

        self.layermap = layermap
        self.lib = lib

    def write(self, filename, cellname: str = "out") -> kf.KCell:
        """Write gds.

        Args:
            filepath: gdspath.
            cellname: for top cell.
        """
        c = kf.KCell(cellname, self.lib)

        for layername, layer in self.layermap.items():
            region = getattr(self, layername)
            try:
                c.shapes(self.lib.layer(layer[0], layer[1])).insert(region)
            except TypeError:
                raise ValueError(
                    f"Unexpected type for region {layername!r}: {type(region)}"
                )
        c.write(filename)
        return c

    def __delattr__(self, element):
        setattr(self, element, Region())

    def get_fill(
        self,
        region,
        size: Tuple[float, float],
        spacing: Tuple[float, float],
        fill_layers: LayerSpecs,
        fill_name: str = "fill",
        fill_cell_name: str = "fill_cell",
    ) -> None:
        """Generates rectangular fill on a set of layers in the region specified.

        Args:
            region: to fill, usually the result of prior boolean operations.
            size: (x,y) dimensions of the fill cell (um).
            spacing: (x,y) spacing of the fill cell (um).
            fill_layers: layers of the fill cell (can be multiple).
            fill_name: fill cell name.
            fill_cell_name: fill cell name.
        """

        fill_cell = kf.KCell(fill_cell_name)
        for layer in fill_layers:
            layer = kf.klib.layer(*layer)
            fill_cell << kf.pcells.waveguide.waveguide(
                width=size[0], length=size[1], layer=layer
            )

        fc_index = fill_cell.cell_index()  # fill cell index
        fc_box = fill_cell.bbox().enlarged(spacing[0] / 2 * 1e3, spacing[1] / 2 * 1e3)
        fill_margin = kf.kdb.Point(0, 0)

        fill = kf.KCell(fill_name)
        return fill.fill_region(
            region, fc_index, fc_box, None, region, fill_margin, None
        )


if __name__ == "__main__":
    from gdsfactory.generic_tech.layer_map import LAYER as l
    import gdsfactory.geometry.maskprep_flat as dp
    import gdsfactory as gf
    import kfactory as kf

    c = gf.Component()
    ring = c << gf.components.coupler_ring()
    floorplan = c << gf.components.bbox(ring.bbox, layer=l.FLOORPLAN)
    c.write_gds("src.gds")

    d = dp.RegionCollection(filepath="src.gds", layermap=dict(l))
    fill_cell = d.get_fill(
        d.FLOORPLAN - d.WG, size=(0.1, 0.1), spacing=(0.1, 0.1), fill_layers=(l.WG,)
    )
    fill_cell.write("fill.gds")
    gf.show("fill.gds")



from __future__ import annotations

from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, LayerSpec


@gf.cell
def grating_coupler_rectangular(
    n_periods: int = 20,
    period: float = 0.75,
    fill_factor: float = 0.5,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    polarization: str = "te",
    wavelength: float = 1.55,
    taper: ComponentSpec = taper_function,
    layer_slab: Optional[LayerSpec] = "SLAB150",
    fiber_angle: float = 15,
    slab_xmin: float = -1.0,
    slab_offset: float = 1.0,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""Grating coupler with rectangular shapes (not elliptical).

    Needs longer taper than elliptical.
    Grating teeth are straight.
    For a focusing grating take a look at grating_coupler_elliptical.

    Args:
        n_periods: number of grating teeth.
        period: grating pitch.
        fill_factor: ratio of grating width vs gap.
        width_grating: 11.
        length_taper: 150.
        polarization: 'te' or 'tm'.
        wavelength: in um.
        taper: function.
        layer_slab: layer that protects the slab under the grating.
        fiber_angle: in degrees.
        slab_xmin: where 0 is at the start of the taper.
        slab_offset: from edge of grating to edge of the slab.
        cross_section: for input waveguide port.
        kwargs: cross_section settings.

    .. code::

        side view
                      fiber

                   /  /  /  /
                  /  /  /  /

                _|-|_|-|_|-|___ layer
                   layer_slab |
            o1  ______________|


        top view     _________
                    /| | | | |
                   / | | | | |
                  /taper_angle
                 /_ _| | | | |
        wg_width |   | | | | |
                 \   | | | | |
                  \  | | | | |
                   \ | | | | |
                    \|_|_|_|_|
                 <-->
                taper_length
    """
    xs = gf.get_cross_section(cross_section, **kwargs)
    wg_width = xs.width
    layer = xs.layer

    c = Component()
    taper_ref = c << gf.get_component(
        taper,
        length=length_taper,
        width2=width_grating,
        width1=wg_width,
        layer=layer,
        cross_section=cross_section,
    )

    c.add_port(port=taper_ref.ports["o1"], name="o1")
    x0 = taper_ref.xmax

    for i in range(n_periods):
        xsize = gf.snap.snap_to_grid(period * fill_factor)
        cgrating = c.add_ref(
            rectangle(size=(xsize, width_grating), layer=layer, port_type=None)
        )
        cgrating.xmin = gf.snap.snap_to_grid(x0 + i * period)
        cgrating.y = 0

    xport = np.round((x0 + cgrating.x) / 2, 3)

    name = f"opt_{polarization.lower()}_{int(wavelength*1e3)}_{int(fiber_angle)}"
    c.add_port(
        name=name,
        port_type=name,
        center=(xport, 0),
        orientation=0,
        width=width_grating,
        layer=layer,
    )
    c.info["polarization"] = polarization
    c.info["wavelength"] = wavelength
    gf.asserts.grating_coupler(c)

    if layer_slab:
        slab_xmin += length_taper
        slab_xsize = cgrating.xmax + slab_offset
        slab_ysize = c.ysize + 2 * slab_offset
        yslab = slab_ysize / 2
        c.add_polygon(
            [
                (slab_xmin, yslab),
                (slab_xsize, yslab),
                (slab_xsize, -yslab),
                (slab_xmin, -yslab),
            ],
            layer_slab,
        )
    if xs.add_bbox:
        c = xs.add_bbox(c)
    if xs.add_pins:
        c = xs.add_pins(c)
    return c


if __name__ == "__main__":
    # c = grating_coupler_rectangular(name='gcu', partial_etch=True)
    # c = grating_coupler_rectangular()
    c = gf.routing.add_fiber_array(grating_coupler=grating_coupler_rectangular)
    print(c.ports)
    c.show(show_ports=True)



from __future__ import annotations

import copy
import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bezier import bezier
from gdsfactory.typings import CrossSectionSpec, Float2
from gdsfactory.config import logger


@cell
def bend_s(
    size: Float2 = (11.0, 2.0),
    npoints: int = 99,
    cross_section: CrossSectionSpec = "strip",
    check_min_radius: bool = False,
    **kwargs,
) -> Component:
    """Return S bend with bezier curve.

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        size: in x and y direction.
        npoints: number of points.
        cross_section: spec.
        check_min_radius: raise ValueError if radius below min_bend_radius.
        kwargs: cross_section settings.
    """
    c = Component()
    dx, dy = size

    bend = bezier(
        control_points=((0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)),
        npoints=npoints,
        cross_section=cross_section,
        **kwargs,
    )
    bend_ref = c << bend
    c.add_ports(bend_ref.ports)
    c.copy_child_info(bend)

    cross_section = gf.get_cross_section(cross_section, **kwargs)
    if (
        cross_section.radius is not None
        and c.info["min_bend_radius"] < cross_section.radius
    ):
        if check_min_radius:
            raise ValueError(
                f"The min bend radius of the generated s bend {c.info['min_bend_radius']} "
                f"is below the bend radius of the waveguide {cross_section.radius}"
            )
        else:
            logger.warning(
                f"The min bend radius of the generated s bend {c.info['min_bend_radius']} is below the bend radius of the waveguide {cross_section.radius}"
            )

    return c


def get_min_sbend_size(
    size: Float2 = (None, 10.0),
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
):
    """
    Returns the minimum sbend size to comply with bend radius requirements.

     Args:
        size: in x and y direction. One of them is None, which is the size we need to figure out.
        cross_section: spec.
        kwargs: cross_section settings.

    """

    cross_section_f = gf.get_cross_section(cross_section, **kwargs)

    if size[0] is None:
        ind = 0
        known_s = size[1]
    elif size[1] is None:
        ind = 1
        known_s = size[0]
    else:
        raise ValueError("One of the two elements in size has to be None")

    min_radius = cross_section_f.radius

    if min_radius is None:
        raise ValueError("The min radius for the specified layer is not known!")

    min_size = np.inf

    # Guess sizes, iterate over them until we cannot achieve the min radius
    # the max size corresponds to an ellipsoid
    max_size = 2.5 * np.sqrt(np.abs(min_radius * known_s))

    sizes = np.linspace(max_size, 0.1 * max_size, 100)

    for i, s in enumerate(sizes):
        sz = copy.deepcopy(size)
        sz[ind] = s
        # print(sz)
        try:
            bend_s(
                size=sz, cross_section=cross_section, check_min_radius=True, **kwargs
            )
            # print(c.info['min_bend_radius'])
            min_size = sizes[i]
        except ValueError:
            break

    return min_size


if __name__ == "__main__":
    c = bend_s(bbox_offsets=[0.5], bbox_layers=[(111, 0)], width=2)
    # c = bend_s(size=[10, 2.5])  # 10um bend radius
    # c = bend_s(size=[20, 3], cross_section="rib")  # 10um bend radius
    # c.pprint()
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    c.show(show_ports=False)



from __future__ import annotations

from typing import Dict, Optional, Tuple

import pygmsh

from gdsfactory.pdk import get_layer_stack
from gdsfactory.technology import LayerStack
from gdsfactory.typings import ComponentOrReference, Layer


def surface_loop_from_vertices(model, xmin, xmax, ymin, ymax, zmin, zmax, resolution):
    """Returns surface loop of prism from bounding box.

    Args:
        xmin: minimal x-value for bounding box
        xmax: maximal x-value for bounding box
        ymin: minimal y-value for bounding box
        ymax: maximal y-value for bounding box
        zmin: minimal z-value for bounding box
        zmax: maximal z-value for bounding box

    """
    channel_surfaces = []
    for coords in [
        [
            [xmin, ymin, zmin],
            [xmin, ymin, zmax],
            [xmin, ymax, zmax],
            [xmin, ymax, zmin],
        ],
        [
            [xmax, ymin, zmin],
            [xmax, ymin, zmax],
            [xmax, ymax, zmax],
            [xmax, ymax, zmin],
        ],
        [
            [xmax, ymin, zmin],
            [xmax, ymin, zmax],
            [xmin, ymin, zmax],
            [xmin, ymin, zmin],
        ],
        [
            [xmax, ymax, zmin],
            [xmax, ymax, zmax],
            [xmin, ymax, zmax],
            [xmin, ymax, zmin],
        ],
        [
            [xmin, ymin, zmin],
            [xmin, ymax, zmin],
            [xmax, ymax, zmin],
            [xmax, ymin, zmin],
        ],
        [
            [xmin, ymin, zmax],
            [xmin, ymax, zmax],
            [xmax, ymax, zmax],
            [xmax, ymin, zmax],
        ],
    ]:
        points = [model.add_point(coord, mesh_size=resolution) for coord in coords]
        channel_lines = [
            model.add_line(points[i], points[i + 1]) for i in range(-1, len(points) - 1)
        ]
        channel_loop = model.add_curve_loop(channel_lines)
        channel_surfaces.append(model.add_plane_surface(channel_loop))
    surface_loop = model.add_surface_loop(channel_surfaces)
    return channel_surfaces, surface_loop


def mesh3D(
    component: ComponentOrReference,
    base_resolution: float = 0.2,
    refine_resolution: Optional[Dict[Layer, float]] = None,
    padding: Tuple[float, float, float, float, float, float] = (
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
    ),
    layer_stack: Optional[LayerStack] = None,
    exclude_layers: Optional[Tuple[Layer, ...]] = None,
):
    """Returns gmsh 3D geometry of component.

    Similar to component.to_3d(), but returns a gmsh mesh with:
    - layer-dependent mesh resolution
    - different physical blocks for different objects
    - sub-object labels for introduction in physical solvers (e.g. edges)

    Args:
        component: Component or ComponentReference.
        base_resolution: background mesh resolution (um).
        refine_resolution: feature mesh resolution (um); layer dependent via a dict (default to base_resolution).
        padding: amount (west, east, south, north, down, up) to enlarge simulation region beyond features (um).

    """
    layer_stack = layer_stack or get_layer_stack()
    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    exclude_layers = exclude_layers or ()

    geometry = pygmsh.occ.geometry.Geometry()

    model = geometry.__enter__()

    zmin_cell = 0  # np.inf
    zmax_cell = 1.22  # -np.inf

    bbox = component.bbox
    xmin_cell = bbox[0][0] - padding[0]
    ymin_cell = bbox[0][1] - padding[2]
    xmax_cell = bbox[1][0] + padding[1]
    ymax_cell = bbox[1][1] + padding[3]

    # Create element resolution dict
    # refine_dict = {
    #     layer: refine_resolution[layer]
    #     if layer in refine_resolution.keys()
    #     else base_resolution
    #     for layer in component.get_layers()
    # }

    # Features
    # blocks = []
    # for layer, polygons in component.get_polygons(by_spec=True).items():
    for layer in component.layers:
        if (
            layer not in exclude_layers
            and layer in layer_to_thickness
            and layer in layer_to_zmin
        ):
            height = layer_to_thickness[layer]
            zmin_layer = layer_to_zmin[layer]
            zmax_layer = zmin_layer + height

            if zmin_layer < zmin_cell:
                zmin_cell = zmin_layer
            if zmax_layer > zmax_cell:
                zmax_cell = zmax_layer

            # num_layers = int(height/refine_dict[layer])
            # i = 0
            # for polygon in polygons:
            #     points = [model.add_point([polygon_point[0], polygon_point[1], zmin_layer], mesh_size=refine_dict[layer]) for polygon_point in polygon]
            #     polygon_lines = [
            #         model.add_line(points[i], points[i + 1]) for i in range(-1, len(points) - 1)
            #     ]
            #     polygon_loop = model.add_curve_loop(polygon_lines)
            #     polygon_surface = model.add_plane_surface(polygon_loop)
            #     polygon_top, polygon_volume, polygon_lat = model.extrude(polygon_surface, [0,0,height], num_layers=int(height/refine_dict[layer]))
            #     model.add_physical(polygon_volume, f"{layer}_{i}")
            #     blocks.append("{layer}_{i}")
            #     i += 1

    zmin_cell -= padding[4]
    zmax_cell += padding[5]

    # Background oxide
    # Generate boundary surfaces
    # cell_planes = []
    # for coords in [
    #                 [
    #                     [xmin_cell,ymin_cell,zmin_cell],
    #                     [xmin_cell,ymin_cell,zmax_cell],
    #                     [xmin_cell,ymax_cell,zmax_cell],
    #                     [xmin_cell,ymax_cell,zmin_cell]
    #                 ],
    #                 [
    #                     [xmax_cell,ymin_cell,zmin_cell],
    #                     [xmax_cell,ymin_cell,zmax_cell],
    #                     [xmax_cell,ymax_cell,zmax_cell],
    #                     [xmax_cell,ymax_cell,zmin_cell]
    #                 ],
    #                 [
    #                     [xmax_cell,ymin_cell,zmin_cell],
    #                     [xmax_cell,ymin_cell,zmax_cell],
    #                     [xmin_cell,ymin_cell,zmax_cell],
    #                     [xmin_cell,ymin_cell,zmin_cell]
    #                 ],
    #                 [
    #                     [xmax_cell,ymax_cell,zmin_cell],
    #                     [xmax_cell,ymax_cell,zmax_cell],
    #                     [xmin_cell,ymax_cell,zmax_cell],
    #                     [xmin_cell,ymax_cell,zmin_cell]
    #                 ],
    #                 [
    #                     [xmin_cell, ymin_cell, zmin_cell],
    #                     [xmin_cell, ymax_cell, zmin_cell],
    #                     [xmax_cell, ymax_cell, zmin_cell],
    #                     [xmax_cell, ymin_cell, zmin_cell]
    #                 ],
    #                 [
    #                     [xmin_cell, ymin_cell, zmax_cell],
    #                     [xmin_cell, ymax_cell, zmax_cell],
    #                     [xmax_cell, ymax_cell, zmax_cell],
    #                     [xmax_cell, ymin_cell, zmax_cell]
    #                 ],
    #             ]:
    #     points = []
    #     for coord in coords:
    #         points.append(model.add_point(coord, mesh_size=base_resolution))
    #     channel_lines = [
    #         model.add_line(points[i], points[i + 1]) for i in range(-1, len(points) - 1)
    #     ]
    #     channel_loop = model.add_curve_loop(channel_lines)
    #     plane_surface = model.add_plane_surface(channel_loop)
    #     cell_planes.append(plane_surface)

    channel_surfaces, surface_loop = surface_loop_from_vertices(
        model,
        xmin=xmin_cell,
        xmax=xmax_cell,
        ymin=ymin_cell,
        ymax=ymax_cell,
        zmin=0,
        zmax=1,
        resolution=0.1,
    )
    # i = 0

    cell_volume = model.add_volume(surface_loop)  # , holes=blocks)
    # for surface in channel_surfaces:
    #     geometry.boolean_fragments([cell_volume, surface], [], delete_first=True, delete_other=False)
    # model.add_physical(surface, f"oxide_{i}")
    # i += 1
    # cell_volume = model.boolean_difference(cell_volume, channel_surfaces, delete_first=True, delete_other=False)
    cell_volume = model.boolean_fragments(
        cell_volume, channel_surfaces, delete_first=True, delete_other=True
    )
    # model.add_physical(cell_volume, "oxide_vol")

    # for coords in [[
    #                 [xmin_cell, ymin_cell, zmin_cell],
    #                 [xmin_cell, ymax_cell, zmin_cell],
    #                 [xmax_cell, ymax_cell, zmin_cell],
    #                 [xmax_cell, ymin_cell, zmin_cell]
    #             ]]:
    #     points = []
    #     for coord in coords:
    #         points.append(model.add_point(coord, mesh_size=base_resolution))

    #     channel_lines = [
    #         model.add_line(points[i], points[i + 1]) for i in range(-1, len(points) - 1)
    #     ]
    #     channel_loop = model.add_curve_loop(channel_lines)
    #     plane_surface = model.add_plane_surface(channel_loop)

    # top, volume, lat = model.extrude(plane_surface, [0,0,zmax_cell-zmin_cell], num_layers=int((zmax_cell-zmin_cell)/base_resolution))
    # bottom = plane_surface

    # model.add_physical(cell_volume, "oxide")
    # model.add_physical(lat, "lat")
    # model.add_physical(top, "top")
    # model.add_physical(bottom, "bottom")
    geometry.generate_mesh(dim=3, verbose=True)

    return geometry


if __name__ == "__main__":
    import gdsfactory as gf

    heaters = gf.Component("heaters")
    heater1 = gf.components.straight(length=2)
    heater2 = gf.components.straight(length=2).move([0, 1])

    heaters = gf.Component()
    heaters << heater1
    heaters << heater2
    heaters.show()

    print(heaters.get_layers())

    geometry = mesh3D(
        heaters,
        exclude_layers=[(1, 10)],
        refine_resolution={(1, 0): 0.05, (47, 0): 0.05},
    )

    import gmsh

    gmsh.write("mesh.msh")
    gmsh.clear()
    geometry.__exit__()

    import meshio

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=False):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    # line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    # meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    meshio.write("mesh.xdmf", triangle_mesh)

    # for layer, polygons in heaters.get_polygons(by_spec=True).items():
    #     print(layer, polygons)



from gdsfactory.generic_tech.layer_map import LAYER
from gdsfactory.technology import LayerLevel, LayerStack

nm = 1e-3


def get_layer_stack(
    thickness_wg: float = 220 * nm,
    thickness_slab_deep_etch: float = 90 * nm,
    thickness_slab_shallow_etch: float = 150 * nm,
    sidewall_angle_wg: float = 10,
    thickness_clad: float = 3.0,
    thickness_nitride: float = 350 * nm,
    thickness_ge: float = 500 * nm,
    gap_silicon_to_nitride: float = 100 * nm,
    zmin_heater: float = 1.1,
    zmin_metal1: float = 1.1,
    thickness_metal1: float = 700 * nm,
    zmin_metal2: float = 2.3,
    thickness_metal2: float = 700 * nm,
    zmin_metal3: float = 3.2,
    thickness_metal3: float = 2000 * nm,
    substrate_thickness: float = 10.0,
    box_thickness: float = 3.0,
    undercut_thickness: float = 5.0,
) -> LayerStack:
    """Returns generic LayerStack.

    based on paper https://www.degruyter.com/document/doi/10.1515/nanoph-2013-0034/html

    Args:
        thickness_wg: waveguide thickness in um.
        thickness_slab_deep_etch: for deep etched slab.
        thickness_shallow_etch: thickness for the etch.
        sidewall_angle_wg: waveguide side angle.
        thickness_clad: cladding thickness in um.
        thickness_nitride: nitride thickness in um.
        thickness_ge: germanium thickness.
        gap_silicon_to_nitride: distance from silicon to nitride in um.
        zmin_heater: TiN heater.
        zmin_metal1: metal1.
        thickness_metal1: metal1 thickness.
        zmin_metal2: metal2.
        thickness_metal2: metal2 thickness.
        zmin_metal3: metal3.
        thickness_metal3: metal3 thickness.
        substrate_thickness: substrate thickness in um.
        box_thickness: bottom oxide thickness in um.
        undercut_thickness: thickness of the silicon undercut.
    """

    thickness_deep_etch = thickness_wg - thickness_slab_deep_etch
    thickness_shallow_etch = thickness_wg - thickness_slab_shallow_etch

    class GenericLayerStack(LayerStack):
        substrate = LayerLevel(
            layer=LAYER.WAFER,
            thickness=substrate_thickness,
            zmin=-substrate_thickness - box_thickness,
            material="si",
            mesh_order=99,
        )
        box = LayerLevel(
            layer=LAYER.WAFER,
            thickness=box_thickness,
            zmin=-box_thickness,
            material="sio2",
            mesh_order=99,
        )
        core = LayerLevel(
            layer=LAYER.WG,
            thickness=thickness_wg,
            zmin=0.0,
            material="si",
            mesh_order=2,
            sidewall_angle=sidewall_angle_wg,
            width_to_z=0.5,
        )
        shallow_etch = LayerLevel(
            layer=LAYER.SHALLOW_ETCH,
            thickness=thickness_shallow_etch,
            zmin=0.0,
            material="si",
            mesh_order=1,
            layer_type="etch",
            into=["core"],
            derived_layer=LAYER.SLAB150,
        )
        deep_etch = LayerLevel(
            layer=LAYER.DEEP_ETCH,
            thickness=thickness_deep_etch,
            zmin=0.0,
            material="si",
            mesh_order=1,
            layer_type="etch",
            into=["core"],
            derived_layer=LAYER.SLAB90,
        )
        clad = LayerLevel(
            # layer=LAYER.WGCLAD,
            layer=LAYER.WAFER,
            zmin=0.0,
            material="sio2",
            thickness=thickness_clad,
            mesh_order=10,
        )
        slab150 = LayerLevel(
            layer=LAYER.SLAB150,
            thickness=150e-3,
            zmin=0,
            material="si",
            mesh_order=3,
        )
        slab90 = LayerLevel(
            layer=LAYER.SLAB90,
            thickness=thickness_slab_deep_etch,
            zmin=0.0,
            material="si",
            mesh_order=2,
        )
        nitride = LayerLevel(
            layer=LAYER.WGN,
            thickness=thickness_nitride,
            zmin=thickness_wg + gap_silicon_to_nitride,
            material="sin",
            mesh_order=2,
        )
        ge = LayerLevel(
            layer=LAYER.GE,
            thickness=thickness_ge,
            zmin=thickness_wg,
            material="ge",
            mesh_order=1,
        )
        undercut = LayerLevel(
            layer=LAYER.UNDERCUT,
            thickness=-undercut_thickness,
            zmin=-box_thickness,
            material="air",
            z_to_bias=[
                [0, 0.3, 0.6, 0.8, 0.9, 1],
                [-0, -0.5, -1, -1.5, -2, -2.5],
            ],
            mesh_order=1,
        )
        via_contact = LayerLevel(
            layer=LAYER.VIAC,
            thickness=zmin_metal1 - thickness_slab_deep_etch,
            zmin=thickness_slab_deep_etch,
            material="Aluminum",
            mesh_order=1,
            sidewall_angle=-10,
            width_to_z=0,
        )
        metal1 = LayerLevel(
            layer=LAYER.M1,
            thickness=thickness_metal1,
            zmin=zmin_metal1,
            material="Aluminum",
            mesh_order=2,
        )
        heater = LayerLevel(
            layer=LAYER.HEATER,
            thickness=750e-3,
            zmin=zmin_heater,
            material="TiN",
            mesh_order=1,
        )
        via1 = LayerLevel(
            layer=LAYER.VIA1,
            thickness=zmin_metal2 - (zmin_metal1 + thickness_metal1),
            zmin=zmin_metal1 + thickness_metal1,
            material="Aluminum",
            mesh_order=2,
        )
        metal2 = LayerLevel(
            layer=LAYER.M2,
            thickness=thickness_metal2,
            zmin=zmin_metal2,
            material="Aluminum",
            mesh_order=2,
        )
        via2 = LayerLevel(
            layer=LAYER.VIA2,
            thickness=zmin_metal3 - (zmin_metal2 + thickness_metal2),
            zmin=zmin_metal2 + thickness_metal2,
            material="Aluminum",
            mesh_order=1,
        )
        metal3 = LayerLevel(
            layer=LAYER.M3,
            thickness=thickness_metal3,
            zmin=zmin_metal3,
            material="Aluminum",
            mesh_order=2,
        )

    return GenericLayerStack()


LAYER_STACK = get_layer_stack()

if __name__ == "__main__":
    # ls = get_layer_stack(substrate_thickness=50.0)
    ls = get_layer_stack()
    script = ls.get_klayout_3d_script()
    print(script)
    # print(ls.get_layer_to_material())
    # print(ls.get_layer_to_thickness())



import numpy as np

import gdsfactory as gf
from gdsfactory.generic_tech import LAYER_STACK
from gdsfactory.simulation.fem.mode_solver import compute_cross_section_modes
from gdsfactory.technology import LayerStack
from femwell import mode_solver


NUM_MODES = 1

PDK = gf.get_generic_pdk()
PDK.activate()


def compute_modes(
    overwrite: bool = True, with_cache: bool = False, num_modes: int = NUM_MODES
):
    filtered_layerstack = LayerStack(
        layers={
            k: LAYER_STACK.layers[k]
            for k in (
                "core",
                "clad",
                "slab90",
                "box",
            )
        }
    )

    filtered_layerstack.layers["core"].thickness = 0.2

    resolutions = {
        "core": {"resolution": 0.02, "distance": 2},
        "clad": {"resolution": 0.2, "distance": 1},
        "box": {"resolution": 0.2, "distance": 1},
        "slab90": {"resolution": 0.05, "distance": 1},
    }
    lams, basis, xs = compute_cross_section_modes(
        cross_section="rib",
        layerstack=filtered_layerstack,
        wavelength=1.55,
        num_modes=num_modes,
        order=1,
        radius=np.inf,
        resolutions=resolutions,
        overwrite=overwrite,
        with_cache=with_cache,
    )
    return lams, basis, xs


def test_compute_cross_section_mode():
    lams, basis, xs = compute_modes()
    assert len(lams) == NUM_MODES, len(lams)


def test_compute_cross_section_mode_cache():
    # write mode in cache
    lams, basis, xs = compute_modes(with_cache=True, overwrite=False)

    # load mode from cache
    lams, basis, xs = compute_modes(with_cache=True, overwrite=False)
    mode_solver.plot_mode(
        basis=basis,
        mode=np.real(xs[0]),
        plot_vectors=False,
        colorbar=True,
        title="E",
        direction="y",
    )


if __name__ == "__main__":
    test_compute_cross_section_mode()
    # test_compute_cross_section_mode_cache()
    # lams, basis, xs = compute_modes(with_cache=True, overwrite=False)
    # mode_solver.plot_mode(
    #     basis=basis,
    #     mode=np.real(xs[0]),
    #     plot_vectors=False,
    #     colorbar=True,
    #     title="E",
    #     direction="y",
    # )



from gdsfactory.pdk import get_layer_stack
from gdsfactory.simulation.eme.meow_eme import MEOW
from gdsfactory.simulation.sax.build_model import Model


class MeowEMEModel(Model):
    def __init__(self, **kwargs) -> None:
        """Generic model inferred from MEOW EME calculations."""
        super().__init__(**kwargs)
        return None

    def get_Sparameters(self, input_dict, output_vector_labels):
        """For EME, results are directly S-parameters."""
        return self.get_results(self, input_dict)

    def get_results(self, input_dict):
        """Setup and run a simulation with one set of inputs."""
        param_dict, layerstack_param_dict = self.parse_input_dict(input_dict)
        input_component = self.component(param_dict)
        input_layerstack = self.perturb_layerstack(layerstack_param_dict)

        eme = MEOW(
            component=input_component,
            layerstack=input_layerstack,
            wavelength=float(input_dict["wavelength"])
            if "wavelength" in input_dict
            else 1.55,
            num_modes=int(self.simulation_settings["num_eme_modes"])
            if "num_eme_modes" in input_dict
            else 4,
            cell_length=float(self.simulation_settings["cell_length"])
            if "cell_length" in input_dict
            else 0.5,
            resolution_x=int(self.simulation_settings["mode_res"])
            if "mode_res" in input_dict
            else 100,
            resolution_y=int(self.simulation_settings["mode_res"])
            if "mode_res" in input_dict
            else 100,
            spacing_x=float(self.simulation_settings["spacing_x"])
            if "spacing_x" in input_dict
            else 1,
            spacing_y=float(self.simulation_settings["spacing_y"])
            if "spacing_y" in input_dict
            else 1,
            overwrite=self.simulation_settings["overwrite"]
            if "overwrite" in input_dict
            else False,
        )
        return eme.compute_sparameters()


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.cross_section import rib, strip
    from gdsfactory.simulation.sax.parameter import LayerStackThickness, NamedParameter
    from gdsfactory.technology import LayerStack

    PDK = gf.get_generic_pdk()
    PDK.activate()

    c = gf.components.taper_cross_section_linear(
        cross_section1=rib(width=2), cross_section2=rib(width=0.5)
    )
    c.show()

    # layerstack = gf.tech.get_layer_stack_generic()

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "box",
                "clad",
            )
        }
    )

    def trainable_strip_strip_taper(parameters):
        return gf.components.taper_cross_section_linear(
            length=parameters["length"],
            cross_section1=strip(width=0.5),
            cross_section2=strip(width=parameters["width2"]),
        )

    strip_strip_taper_model = MeowEMEModel(
        component=trainable_strip_strip_taper,
        layerstack=filtered_layerstack,
        simulation_settings={
            "mode_res": 100,
            "cell_length": 0.5,
            "num_eme_modes": 4,
            "spacing_x": 1,
            "spacing_y": -3,
            "overwrite": False,
        },
        trainable_parameters={
            "length": NamedParameter(min_value=5, max_value=6, nominal_value=6, step=1),
            "width2": NamedParameter(
                min_value=1.0, max_value=1.0, nominal_value=1.0, step=1.0
            ),
            "wavelength": NamedParameter(
                min_value=1.545, max_value=1.555, nominal_value=1.55, step=0.005
            ),
            "core_thickness": LayerStackThickness(
                layerstack=filtered_layerstack,
                min_value=0.19,
                max_value=0.25,
                nominal_value=0.22,
                layername="core",
                step=0.3,
            ),
        },
        num_modes=2,
    )

    # input_vectors, output_vectors = strip_strip_taper_model.get_data()
    interpolator = strip_strip_taper_model.get_nd_nd_interp()

    import jax.numpy as jnp

    params = jnp.stack(
        jnp.broadcast_arrays(
            jnp.asarray([6.0, 6.0, 6.0]),
            jnp.asarray([1.0, 1.0, 1.0]),
            jnp.asarray([1.55, 1.5, 1.55]),
            jnp.asarray([0.22, 0.22, 0.19]),
        ),
        0,
    )

    print(interpolator["o1@0,o2@0"](params))



from __future__ import annotations

from collections import OrderedDict
from itertools import combinations, product
from typing import Dict, Optional

import gmsh
import meshio
import numpy as np
import pygmsh
from scipy.interpolate import NearestNDInterpolator
from shapely.geometry import LineString, Point, Polygon

from gdsfactory.simulation.gmsh.break_geometry import break_geometry
from gdsfactory.simulation.gmsh.meshtracker import MeshTracker
from gdsfactory.simulation.gmsh.refine import (
    global_callback_refinement,
    surface_interface_refinement,
)


def define_entities(model, shapes_dict: OrderedDict, atol=1e-3):
    """Adds the polygons and lines from a "shapes_dict" as physical entities in the pygmsh model "model".

    Args:
        model: pygmsh model to define the entities into.
        shapes_dict: OrderedDict containing shapes to mesh as values, and keys as physical tags.

    Returns:
        model: updated pygmsh model.
        meshtracker: meshtracker object containing the mapping between labels, shapely objects, and gmsh objects.
    """
    # Break up lines and polygon edges so that plane is tiled with no partially overlapping line segments
    polygons_broken_dict, lines_broken_dict = break_geometry(shapes_dict)
    # Instantiate shapely to gmsh map
    meshtracker = MeshTracker(model=model, atol=atol)
    # Add lines, reusing line segments
    model, meshtracker = add_lines(model, meshtracker, lines_broken_dict)
    # Add surfaces, reusing lines
    model, meshtracker = add_surfaces(model, meshtracker, polygons_broken_dict)
    # Add interfaces
    model, meshtracker = tag_interfaces(model, meshtracker, polygons_broken_dict)
    # Synchronize
    model.synchronize()

    return model, meshtracker


def add_lines(model, meshtracker, lines_broken_dict):
    """Add lines, reusing segments via meshtracker.

    Args:
        model: pygmsh model
        meshtracker: meshtracker object
        lines_broken_dict: dict

    Returns:
        model: updated pygmsh model
        meshtracker: updated meshtracker object
    """
    for line_name, line in lines_broken_dict.items():
        meshtracker.add_get_xy_line(line, line_name)

    return model, meshtracker


def add_surfaces(model, meshtracker, polygons_broken_dict):
    """Add surfaces of polygons_broken_dict to model, reusing lines via meshtracker.

    Args:
        model: pygmsh model
        meshtracker: meshtracker object
        polygons_broken_dict: dict of

    Returns:
        model: updated pygmsh model
        meshtracker: updated meshtracker object
    """
    for polygon_name, polygons in polygons_broken_dict.items():
        gmsh_surfaces = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            gmsh_surface = meshtracker.add_xy_surface(polygon, f"{polygon_name}")
            gmsh_surfaces.append(gmsh_surface)
        meshtracker.model.add_physical(gmsh_surfaces, f"{polygon_name}")

    return model, meshtracker


def tag_interfaces(model, meshtracker, polygons_broken_dict):
    """Tag all interfacial lines between polygons as logged in meshtracker.

    Args:
        model: pygmsh model
        meshtracker: meshtracker object
        polygons_broken_dict: dict of

    Returns:
        model: updated pygmsh model
        meshtracker: updated meshtracker object
    """
    for surface1, surface2 in combinations(polygons_broken_dict.keys(), 2):
        if interfaces := [
            line
            for index, line in enumerate(meshtracker.gmsh_xy_segments)
            if (
                meshtracker.xy_segments_main_labels[index] == surface1
                and meshtracker.xy_segments_secondary_labels[index] == surface2
            )
            or (
                meshtracker.xy_segments_main_labels[index] == surface2
                and meshtracker.xy_segments_secondary_labels[index] == surface1
            )
        ]:
            model.add_physical(interfaces, f"{surface1}___{surface2}")

    return model, meshtracker


def mesh_from_polygons(
    shapes_dict: OrderedDict,
    resolutions: Optional[Dict[str, float]] = None,
    mesh_scaling_factor: float = 1.0,
    default_resolution_min: float = 0.01,
    default_resolution_max: float = 0.5,
    filename: Optional[str] = None,
    global_meshsize_array: Optional[np.array] = None,
    global_meshsize_interpolant_func: Optional[callable] = NearestNDInterpolator,
    verbosity: Optional[bool] = False,
    atol: Optional[float] = 1e-4,
):
    """Return a 2D mesh from an ordered dict of shapely polygons.

    Args:
        shapes_dict: OrderedDict containing shapes to mesh.
        resolutions: Dict with same keys as shapes_dict. containing another dict with resolution settings:
            resolution: density of the mesh
            distance: "dropoff" distance of this resolution away from the surface
        default_resolution_min: gmsh default smallest characteristic length
        default_resolution_max: gmsh default largest characteristic length
        filename: to save the mesh
        global_meshsize_array: array [x,y,z,lc] defining local mesh sizes. Not used if None
        global_meshsize_interpolant: interpolation function for array [x,y,z,lc]. Default scipy.interpolate.NearestNDInterpolator
        verbosity: boolean, gmsh stdout as it meshes
        atol: tolerance used to establish equivalency between vertices
    """
    global_meshsize_callback_bool = global_meshsize_array is not None

    geometry = pygmsh.occ.geometry.Geometry()
    model = geometry.__enter__()

    geometry.characteristic_length_min = default_resolution_min
    geometry.characteristic_length_max = default_resolution_max

    # Define geometry
    (
        model,
        meshtracker,
    ) = define_entities(model, shapes_dict, atol)

    # Synchronize
    model.synchronize()

    # Refinement
    if not global_meshsize_callback_bool and resolutions:
        surface_interface_refinement(
            model,
            meshtracker,
            resolutions,
            default_resolution_min,
            default_resolution_max,
        )

    elif global_meshsize_callback_bool:
        global_callback_refinement(
            model,
            global_meshsize_array,
            global_meshsize_interpolant_func,
        )

    # HACK: force shared nodes across interfaces
    gmsh.model.occ.remove_all_duplicates()

    # Perform meshing
    gmsh.option.setNumber("Mesh.ScalingFactor", mesh_scaling_factor)
    mesh = geometry.generate_mesh(dim=2, verbose=verbosity)

    if filename:
        gmsh.write(f"{filename}")

    return mesh


def create_physical_mesh(mesh, cell_type, prune_z=True):
    cells = mesh.get_cells_type(cell_type)
    cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
    points = mesh.points
    return meshio.Mesh(
        points=points,
        cells={cell_type: cells},
        cell_data={"name_to_read": [cell_data]},
    )


if __name__ == "__main__":
    wmode = 1
    wsim = 2
    hclad = 2
    hbox = 2
    wcore = 0.5
    hcore = 0.22
    offset_core = -0.1
    offset_core2 = 1

    # Lines can be added, which is useful to define boundary conditions at various simulation edges
    left_edge = LineString(
        [Point(-wsim / 2, -hcore / 2 - hbox), Point(-wsim / 2, -hcore / 2 + hclad)]
    )
    right_edge = LineString(
        [Point(wsim / 2, -hcore / 2 - hbox), Point(wsim / 2, -hcore / 2 + hclad)]
    )
    top_edge = LineString(
        [Point(-wsim / 2, -hcore / 2 + hclad), Point(wsim / 2, -hcore / 2 + hclad)]
    )
    bottom_edge = LineString(
        [Point(-wsim / 2, -hcore / 2 - hbox), Point(wsim / 2, -hcore / 2 - hbox)]
    )

    # Polygons not only have an edge, but an interior
    core = Polygon(
        [
            Point(-wcore / 2, -hcore / 2 + offset_core),
            Point(-wcore / 2, hcore / 2 + offset_core),
            Point(wcore / 2, hcore / 2 + offset_core),
            Point(wcore / 2, -hcore / 2 + offset_core),
        ]
    )
    core2 = Polygon(
        [
            Point(-wcore / 2, -hcore / 2 + offset_core2),
            Point(-wcore / 2, hcore / 2 + offset_core2),
            Point(wcore / 2, hcore / 2 + offset_core2),
            Point(wcore / 2, -hcore / 2 + offset_core2),
        ]
    )
    clad = Polygon(
        [
            Point(-wsim / 2, -hcore / 2),
            Point(-wsim / 2, -hcore / 2 + hclad),
            Point(wsim / 2, -hcore / 2 + hclad),
            Point(wsim / 2, -hcore / 2),
        ]
    )
    box = Polygon(
        [
            Point(-wsim / 2, -hcore / 2),
            Point(-wsim / 2, -hcore / 2 - hbox),
            Point(wsim / 2, -hcore / 2 - hbox),
            Point(wsim / 2, -hcore / 2),
        ]
    )

    # The order in which objects are inserted into the OrderedDict determines overrides
    shapes = OrderedDict()
    shapes["left_edge"] = left_edge
    shapes["right_edge"] = right_edge
    shapes["top_edge"] = top_edge
    shapes["bottom_edge"] = bottom_edge
    shapes["core"] = core
    shapes["core2"] = core2
    shapes["clad"] = clad
    shapes["box"] = box

    # The resolution dict is not ordered, and can be used to set mesh resolution at various element
    # The edge of a polygon and another polygon (or entire simulation domain) will form a line object that can be refined independently
    resolutions = {}
    resolutions["core"] = {"resolution": 0.05, "distance": 0}
    resolutions["core___clad"] = {"resolution": 0.01, "distance": 0.5}
    resolutions["clad___box"] = {"resolution": 0.01, "distance": 0.5}
    resolutions["bottom_edge"] = {"resolution": 0.05, "distance": 0.5}
    resolutions["left_edge"] = {"resolution": 0.05, "distance": 0.5}
    # resolutions["clad"] = {"resolution": 0.1, "dist_min": 0.01, "dist_max": 0.3}

    xs = np.linspace(-wsim / 2, wsim / 2, 500)
    ys = np.linspace(-hcore / 2 - hbox, -hcore / 2 + hclad, 500)
    global_meshsize_array = []

    ls0 = 0.02
    ls1 = 0.05
    ls2 = 0.005

    r = 0.5
    for x, y in product(xs, ys):
        if np.abs(x) ** 2 + np.abs(y) ** 2 <= r**2:
            global_meshsize_array.append([x, y, 0, ls2])
        elif y > 0:
            global_meshsize_array.append([x, y, 0, ls1])
        else:
            global_meshsize_array.append([x, y, 0, ls0])

    global_meshsize_array = np.array(global_meshsize_array)

    mesh = mesh_from_polygons(
        shapes,
        resolutions,
        mesh_scaling_factor=0.1,
        filename="mesh.msh",
        global_meshsize_array=global_meshsize_array,
        verbosity=True,
    )

    # gmsh.write("mesh.msh")
    # gmsh.clear()
    # mesh.__exit__()

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=True):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    meshio.write("mesh.xdmf", triangle_mesh)
